"""
This state module is used to manage Wordpress installations

:depends: wp binary from http://wp-cli.org/
"""


def __virtual__():
    if "wordpress.show_plugin" in __salt__:
        return True
    return (False, "wordpress module could not be loaded")


def installed(name, user, admin_user, admin_password, admin_email, title, url):
    """
    Run the initial setup of wordpress

    name
        path to the wordpress installation

    user
        user that owns the files for the wordpress installation

    admin_user
        username for wordpress website administrator user

    admin_password
        password for wordpress website administrator user

    admin_email
        email for wordpress website administrator user

    title
        title for the wordpress website

    url
        url for the wordpress website

    .. code-block:: yaml

        /var/www/html:
          wordpress.installed:
            - title: Daniel's Awesome Blog
            - user: apache
            - admin_user: dwallace
            - admin_email: dwallace@example.com
            - admin_password: password123
            - url: https://blog.dwallace.com
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": False}

    check = __salt__["wordpress.is_installed"](name, user)

    if check:
        ret["result"] = True
        ret["comment"] = f"Wordpress is already installed: {name}"
        return ret
    elif __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"Wordpress will be installed: {name}"
        return ret

    resp = __salt__["wordpress.install"](
        name, user, admin_user, admin_password, admin_email, title, url
    )
    if resp:
        ret["result"] = True
        ret["comment"] = f"Wordpress Installed: {name}"
        ret["changes"] = {"new": resp}
    else:
        ret["comment"] = f"Failed to install wordpress: {name}"

    return ret


def activated(name, path, user):
    """
    Activate wordpress plugins

    name
        name of plugin to activate

    path
        path to wordpress installation

    user
        user who should own the files in the wordpress installation

    .. code-block:: yaml

        HyperDB:
          wordpress.activated:
            - path: /var/www/html
            - user: apache
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": False}

    check = __salt__["wordpress.show_plugin"](name, path, user)

    if check["status"] == "active":
        ret["result"] = True
        ret["comment"] = f"Plugin already activated: {name}"
        return ret
    elif __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"Plugin will be activated: {name}"
        return ret

    resp = __salt__["wordpress.activate"](name, path, user)
    if resp is True:
        ret["result"] = True
        ret["comment"] = f"Plugin activated: {name}"
        ret["changes"] = {
            "old": check,
            "new": __salt__["wordpress.show_plugin"](name, path, user),
        }
    elif resp is None:
        ret["result"] = True
        ret["comment"] = f"Plugin already activated: {name}"
        ret["changes"] = {
            "old": check,
            "new": __salt__["wordpress.show_plugin"](name, path, user),
        }
    else:
        ret["comment"] = f"Plugin failed to activate: {name}"

    return ret


def deactivated(name, path, user):
    """
    Deactivate wordpress plugins

    name
        name of plugin to deactivate

    path
        path to wordpress installation

    user
        user who should own the files in the wordpress installation

    .. code-block:: yaml

        HyperDB:
          wordpress.deactivated:
            - path: /var/www/html
            - user: apache
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": False}

    check = __salt__["wordpress.show_plugin"](name, path, user)

    if check["status"] == "inactive":
        ret["result"] = True
        ret["comment"] = f"Plugin already deactivated: {name}"
        return ret
    elif __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"Plugin will be deactivated: {name}"
        return ret

    resp = __salt__["wordpress.deactivate"](name, path, user)
    if resp is True:
        ret["result"] = True
        ret["comment"] = f"Plugin deactivated: {name}"
        ret["changes"] = {
            "old": check,
            "new": __salt__["wordpress.show_plugin"](name, path, user),
        }
    elif resp is None:
        ret["result"] = True
        ret["comment"] = f"Plugin already deactivated: {name}"
        ret["changes"] = {
            "old": check,
            "new": __salt__["wordpress.show_plugin"](name, path, user),
        }
    else:
        ret["comment"] = f"Plugin failed to deactivate: {name}"

    return ret



"""
Chronos
========

Proxy minion for managing a Chronos cluster.

Dependencies
------------

- :mod:`chronos execution module (salt.modules.chronos) <salt.modules.chronos>`

Pillar
------

The chronos proxy configuration requires a 'base_url' property that points to
the chronos endpoint:

.. code-block:: yaml

    proxy:
      proxytype: chronos
      base_url: http://my-chronos-master.mydomain.com:4400

.. versionadded:: 2015.8.2
"""

import logging

import salt.utils.http

__proxyenabled__ = ["chronos"]
CONFIG = {}
CONFIG_BASE_URL = "base_url"
log = logging.getLogger(__file__)


def __virtual__():
    return True


def init(opts):
    """
    Perform any needed setup.
    """
    if CONFIG_BASE_URL in opts["proxy"]:
        CONFIG[CONFIG_BASE_URL] = opts["proxy"][CONFIG_BASE_URL]
    else:
        log.error("missing proxy property %s", CONFIG_BASE_URL)
    log.debug("CONFIG: %s", CONFIG)


def ping():
    """
    Is the chronos api responding?
    """
    try:
        response = salt.utils.http.query(
            f"{CONFIG[CONFIG_BASE_URL]}/scheduler/jobs",
            decode_type="json",
            decode=True,
        )
        log.debug(
            "chronos.info returned successfully: %s",
            response,
        )
        if "dict" in response:
            return True
    except Exception as ex:  # pylint: disable=broad-except
        log.error(
            "error pinging chronos with base_url %s: %s",
            CONFIG[CONFIG_BASE_URL],
            ex,
        )
    return False


def shutdown(opts):
    """
    For this proxy shutdown is a no-op
    """
    log.debug("chronos proxy shutdown() called...")



import time

from tests.conftest import FIPS_TESTRUN


def test_reauth(salt_master_factory, event_listener):
    """
    Validate non of our platform need to re-authenticate when runing a job with
    multiprocessing=True.
    """
    sls_name = "issue-64941"
    sls_contents = """
    custom_test_state:
      test.configurable_test_state:
        - name: example
        - changes: True
        - result: True
        - comment: "Nothing has actually been changed"
    """
    events = []

    def handler(data):
        events.append(data)

    event_listener.register_auth_event_handler("test_reauth-master", handler)
    master = salt_master_factory.salt_master_daemon(
        "test_reauth-master",
        overrides={
            "log_level": "info",
            "fips_mode": FIPS_TESTRUN,
            "publish_signing_algorithm": (
                "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
            ),
        },
    )
    sls_tempfile = master.state_tree.base.temp_file(f"{sls_name}.sls", sls_contents)
    minion = master.salt_minion_daemon(
        "test_reauth-minion",
        overrides={
            "log_level": "info",
            "fips_mode": FIPS_TESTRUN,
            "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
            "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
        },
    )
    cli = master.salt_cli()
    start_time = time.time()
    with master.started(), minion.started():
        events = event_listener.get_events(
            [(master.id, "salt/auth")],
            after_time=start_time,
        )
        num_auth = len(events)
        proc = cli.run("state.sls", sls_name, minion_tgt="*")
        assert proc.returncode == 1
        events = event_listener.get_events(
            [(master.id, "salt/auth")],
            after_time=start_time,
        )
        assert num_auth == len(events)



"""
Module for viewing and modifying sysctl parameters
"""

import os
import re

import salt.utils.data
import salt.utils.files
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError

# Define the module's virtual name
__virtualname__ = "sysctl"


def __virtual__():
    """
    Only run on NetBSD systems
    """
    if __grains__["os"] == "NetBSD":
        return __virtualname__
    return (
        False,
        "The netbsd_sysctl execution module failed to load: only available on NetBSD.",
    )


def show(config_file=False):
    """
    Return a list of sysctl parameters for this minion

    config: Pull the data from the system configuration file
        instead of the live data.

    CLI Example:

    .. code-block:: bash

        salt '*' sysctl.show
    """
    roots = (
        "kern",
        "vm",
        "vfs",
        "net",
        "hw",
        "machdep",
        "user",
        "ddb",
        "proc",
        "emul",
        "security",
        "init",
    )
    cmd = "sysctl -ae"
    ret = {}
    out = __salt__["cmd.run"](cmd, output_loglevel="trace")
    comps = [""]
    for line in out.splitlines():
        if any([line.startswith(f"{root}.") for root in roots]):
            comps = re.split("[=:]", line, 1)
            ret[comps[0]] = comps[1]
        elif comps[0]:
            ret[comps[0]] += f"{line}\n"
        else:
            continue
    return ret


def get(name):
    """
    Return a single sysctl parameter for this minion

    CLI Example:

    .. code-block:: bash

        salt '*' sysctl.get hw.physmem
    """
    cmd = f"sysctl -n {name}"
    out = __salt__["cmd.run"](cmd, python_shell=False)
    return out


def assign(name, value):
    """
    Assign a single sysctl parameter for this minion

    CLI Example:

    .. code-block:: bash

        salt '*' sysctl.assign net.inet.icmp.icmplim 50
    """
    ret = {}
    cmd = f'sysctl -w {name}="{value}"'
    data = __salt__["cmd.run_all"](cmd, python_shell=False)

    if data["retcode"] != 0:
        raise CommandExecutionError("sysctl failed: {}".format(data["stderr"]))
    new_name, new_value = data["stdout"].split(":", 1)
    ret[new_name] = new_value.split(" -> ")[-1]
    return ret


def persist(name, value, config="/etc/sysctl.conf"):
    """
    Assign and persist a simple sysctl parameter for this minion

    CLI Example:

    .. code-block:: bash

        salt '*' sysctl.persist net.inet.icmp.icmplim 50
    """
    nlines = []
    edited = False
    value = str(value)

    # create /etc/sysctl.conf if not present
    if not os.path.isfile(config):
        try:
            with salt.utils.files.fopen(config, "w+"):
                pass
        except OSError:
            msg = "Could not create {0}"
            raise CommandExecutionError(msg.format(config))

    with salt.utils.files.fopen(config, "r") as ifile:
        for line in ifile:
            line = salt.utils.stringutils.to_unicode(line)
            m = re.match(rf"{name}(\??=)", line)
            if not m:
                nlines.append(line)
                continue
            else:
                key, rest = line.split("=", 1)
                if rest.startswith('"'):
                    _, rest_v, rest = rest.split('"', 2)
                elif rest.startswith("'"):
                    _, rest_v, rest = rest.split("'", 2)
                else:
                    rest_v = rest.split()[0]
                    rest = rest[len(rest_v) :]
                if rest_v == value:
                    return "Already set"
                new_line = f"{name}{m.group(1)}{value}{rest}"
                nlines.append(new_line)
                edited = True

    if not edited:
        newline = f"{name}={value}"
        nlines.append(f"{newline}\n")

    with salt.utils.files.fopen(config, "wb") as ofile:
        ofile.writelines(salt.utils.data.encode(nlines))

    assign(name, value)

    return "Updated"



"""
    :codeauthor: Tyler Johnson <tjohnson@saltstack.com>
"""

import io
import textwrap
import urllib

import pytest
import requests

from salt import config
from salt.cloud.clouds import proxmox
from tests.support.mock import ANY, MagicMock, call, patch


@pytest.fixture
def profile():
    return {
        "my_proxmox": {
            "provider": "my_proxmox",
            "image": "local:some_image.tgz",
        }
    }


@pytest.fixture
def provider_config(profile):
    return {
        "my_proxmox": {
            "proxmox": {
                "driver": "proxmox",
                "url": "pve@domain.com",
                "user": "cloud@pve",
                "password": "verybadpass",
                "profiles": profile,
            }
        }
    }


@pytest.fixture
def vm():
    return {
        "profile": "my_proxmox",
        "name": "vm4",
        "driver": "proxmox",
        "technology": "qemu",
        "host": "127.0.0.1",
        "clone": True,
        "ide0": "data",
        "sata0": "data",
        "scsi0": "data",
        "net0": "a=b,c=d",
    }


@pytest.fixture
def configure_loader_modules(profile, provider_config):
    return {
        proxmox: {
            "__utils__": {
                "cloud.fire_event": MagicMock(),
                "cloud.filter_event": MagicMock(),
                "cloud.bootstrap": MagicMock(),
            },
            "__opts__": {
                "sock_dir": True,
                "transport": True,
                "providers": provider_config,
                "profiles": profile,
            },
            "__active_provider_name__": "my_proxmox:proxmox",
        }
    }


def test__stringlist_to_dictionary():
    result = proxmox._stringlist_to_dictionary("")
    assert result == {}

    result = proxmox._stringlist_to_dictionary(
        "foo=bar, ignored_space=bar,internal space=bar"
    )
    assert result == {"foo": "bar", "ignored_space": "bar", "internal space": "bar"}

    # Negative cases
    pytest.raises(ValueError, proxmox._stringlist_to_dictionary, "foo=bar,foo")
    pytest.raises(
        ValueError,
        proxmox._stringlist_to_dictionary,
        "foo=bar,totally=invalid=assignment",
    )


def test__dictionary_to_stringlist():
    result = proxmox._dictionary_to_stringlist({})
    assert result == ""

    result = proxmox._dictionary_to_stringlist({"a": "a"})
    assert result == "a=a"

    result = proxmox._dictionary_to_stringlist({"a": "a", "b": "b"})
    assert result == "a=a,b=b"


def test__reconfigure_clone_net_hdd(vm):
    # The return_value is for the net reconfigure assertions, it is irrelevant for the rest
    with patch(
        "salt.cloud.clouds.proxmox._get_properties",
        MagicMock(return_value=["net0", "ide0", "sata0", "scsi0"]),
    ), patch.object(
        proxmox, "query", return_value={"net0": "c=overwritten,g=h"}
    ) as query:
        # Test a vm that lacks the required attributes
        proxmox._reconfigure_clone({}, 0)
        query.assert_not_called()

        # Test a fully mocked vm
        proxmox._reconfigure_clone(vm, 0)

        # net reconfigure
        query.assert_any_call("get", "nodes/127.0.0.1/qemu/0/config")
        query.assert_any_call(
            "post", "nodes/127.0.0.1/qemu/0/config", {"net0": "a=b,c=d,g=h"}
        )

        # hdd reconfigure
        query.assert_any_call("post", "nodes/127.0.0.1/qemu/0/config", {"ide0": "data"})
        query.assert_any_call(
            "post", "nodes/127.0.0.1/qemu/0/config", {"sata0": "data"}
        )
        query.assert_any_call(
            "post", "nodes/127.0.0.1/qemu/0/config", {"scsi0": "data"}
        )


def test__reconfigure_clone_params():
    """
    Test cloning a VM with parameters to be reconfigured.
    """
    vmid = 201
    properties = {
        "ide2": "cdrom",
        "sata1": "satatest",
        "scsi0": "bootvol",
        "net0": "model=virtio",
        "agent": "1",
        "args": "argsvalue",
        "balloon": "128",
        "ciuser": "root",
        "cores": "2",
        "description": "desc",
        "memory": "256",
        "name": "new2",
        "onboot": "0",
        "sshkeys": "ssh-rsa ABCDEF user@host\n",
    }
    query_calls = [call("get", f"nodes/myhost/qemu/{vmid}/config")]
    for key, value in properties.items():
        if key == "sshkeys":
            value = urllib.parse.quote(value, safe="")
        query_calls.append(
            call(
                "post",
                f"nodes/myhost/qemu/{vmid}/config",
                {key: value},
            )
        )

    mock_query = MagicMock(return_value="")
    with patch(
        "salt.cloud.clouds.proxmox._get_properties",
        MagicMock(return_value=list(properties.keys())),
    ), patch("salt.cloud.clouds.proxmox.query", mock_query):
        vm_ = {
            "profile": "my_proxmox",
            "driver": "proxmox",
            "technology": "qemu",
            "name": "new2",
            "host": "myhost",
            "clone": True,
            "clone_from": 123,
            "ip_address": "10.10.10.10",
        }
        vm_.update(properties)

        proxmox._reconfigure_clone(vm_, vmid)
        mock_query.assert_has_calls(query_calls, any_order=True)


def test_clone():
    """
    Test that an integer value for clone_from
    """
    mock_query = MagicMock(return_value="")
    with patch(
        "salt.cloud.clouds.proxmox._get_properties", MagicMock(return_value=[])
    ), patch("salt.cloud.clouds.proxmox.query", mock_query):
        vm_ = {
            "technology": "qemu",
            "name": "new2",
            "host": "myhost",
            "clone": True,
            "clone_from": 123,
        }

        # CASE 1: Numeric ID
        result = proxmox.create_node(vm_, ANY)
        mock_query.assert_called_once_with(
            "post",
            "nodes/myhost/qemu/123/clone",
            {"newid": ANY},
        )
        assert result == {"vmid": ANY}

        # CASE 2: host:ID notation
        mock_query.reset_mock()
        vm_["clone_from"] = "otherhost:123"
        result = proxmox.create_node(vm_, ANY)
        mock_query.assert_called_once_with(
            "post",
            "nodes/otherhost/qemu/123/clone",
            {"newid": ANY},
        )
        assert result == {"vmid": ANY}


def test_clone_pool():
    """
    Test that cloning a VM passes the pool parameter if present
    """
    mock_query = MagicMock(return_value="")
    with patch(
        "salt.cloud.clouds.proxmox._get_properties", MagicMock(return_value=[])
    ), patch("salt.cloud.clouds.proxmox.query", mock_query):
        vm_ = {
            "technology": "qemu",
            "name": "new2",
            "host": "myhost",
            "clone": True,
            "clone_from": 123,
            "pool": "mypool",
        }

        result = proxmox.create_node(vm_, ANY)
        mock_query.assert_called_once_with(
            "post",
            "nodes/myhost/qemu/123/clone",
            {"newid": ANY, "pool": "mypool"},
        )
        assert result == {"vmid": ANY}


def test_clone_id():
    """
    Test cloning a VM with a specified vmid.
    """
    next_vmid = 101
    explicit_vmid = 201
    upid = "UPID:myhost:00123456:12345678:9ABCDEF0:qmclone:123:root@pam:"

    def mock_query_response(conn_type, option, post_data=None):
        if conn_type == "get" and option == "cluster/tasks":
            return [{"upid": upid, "status": "OK"}]
        if conn_type == "post" and option.endswith("/clone"):
            return upid
        return None

    mock_wait_for_state = MagicMock(return_value=True)
    with patch(
        "salt.cloud.clouds.proxmox._get_properties",
        MagicMock(return_value=["vmid"]),
    ), patch(
        "salt.cloud.clouds.proxmox._get_next_vmid",
        MagicMock(return_value=next_vmid),
    ), patch(
        "salt.cloud.clouds.proxmox.start", MagicMock(return_value=True)
    ), patch(
        "salt.cloud.clouds.proxmox.wait_for_state", mock_wait_for_state
    ), patch(
        "salt.cloud.clouds.proxmox.query", side_effect=mock_query_response
    ):
        vm_ = {
            "profile": "my_proxmox",
            "driver": "proxmox",
            "technology": "qemu",
            "name": "new2",
            "host": "myhost",
            "clone": True,
            "clone_from": 123,
            "ip_address": "10.10.10.10",
        }

        # CASE 1: No vmid specified in profile (previous behavior)
        proxmox.create(vm_)
        mock_wait_for_state.assert_called_with(
            next_vmid,
            "running",
        )

        # CASE 2: vmid specified in profile
        vm_["vmid"] = explicit_vmid
        proxmox.create(vm_)
        mock_wait_for_state.assert_called_with(
            explicit_vmid,
            "running",
        )


def test_find_agent_ips():
    """
    Test find_agent_ip will return an IP
    """

    with patch(
        "salt.cloud.clouds.proxmox.query",
        return_value={
            "result": [
                {
                    "name": "eth0",
                    "ip-addresses": [
                        {"ip-address": "1.2.3.4", "ip-address-type": "ipv4"},
                        {"ip-address": "2001::1:2", "ip-address-type": "ipv6"},
                    ],
                },
                {
                    "name": "eth1",
                    "ip-addresses": [
                        {"ip-address": "2.3.4.5", "ip-address-type": "ipv4"},
                    ],
                },
                {
                    "name": "dummy",
                },
            ]
        },
    ) as mock_query:
        vm_ = {
            "technology": "qemu",
            "host": "myhost",
            "driver": "proxmox",
            "ignore_cidr": "1.0.0.0/8",
        }

        # CASE 1: Test ipv4 and ignore_cidr
        result = proxmox._find_agent_ip(vm_, ANY)
        mock_query.assert_any_call(
            "get", f"nodes/myhost/qemu/{ANY}/agent/network-get-interfaces"
        )

        assert result == "2.3.4.5"

        # CASE 2: Test ipv6

        vm_["protocol"] = "ipv6"
        result = proxmox._find_agent_ip(vm_, ANY)
        mock_query.assert_any_call(
            "get", f"nodes/myhost/qemu/{ANY}/agent/network-get-interfaces"
        )

        assert result == "2001::1:2"


def test__authenticate_with_custom_port():
    """
    Test the use of a custom port for Proxmox connection
    """
    get_cloud_config_mock = [
        "proxmox.connection.url",
        "9999",
        "fakeuser",
        "secretpassword",
        True,
    ]
    requests_post_mock = MagicMock()
    with patch(
        "salt.config.get_cloud_config_value",
        autospec=True,
        side_effect=get_cloud_config_mock,
    ), patch("requests.post", requests_post_mock):
        proxmox._authenticate()
        requests_post_mock.assert_called_with(
            "https://proxmox.connection.url:9999/api2/json/access/ticket",
            verify=True,
            data={"username": ("fakeuser",), "password": "secretpassword"},
            timeout=120,
        )


def _test__import_api(response):
    """
    Test _import_api recognition of varying Proxmox VE responses.
    """
    requests_get_mock = MagicMock()
    requests_get_mock.return_value.status_code = 200
    requests_get_mock.return_value.text = response
    with patch("requests.get", requests_get_mock):
        proxmox._import_api()
    assert proxmox.api == [{"info": {}}]
    return


def test__import_api_v6():
    """
    Test _import_api handling of a Proxmox VE 6 response.
    """
    response = textwrap.dedent(
        """\
        var pveapi = [
            {
                "info" : {
                }
            }
        ]
        ;
        """
    )
    _test__import_api(response)


def test__import_api_v7():
    """
    Test _import_api handling of a Proxmox VE 7 response.
    """
    response = textwrap.dedent(
        """\
        const apiSchema = [
            {
                "info" : {
                }
            }
        ]
        ;
        """
    )
    _test__import_api(response)


def test__authenticate_success():
    response = requests.Response()
    response.status_code = 200
    response.reason = "OK"
    response.raw = io.BytesIO(
        b"""{"data":{"CSRFPreventionToken":"01234567:dG9rZW4=","ticket":"PVE:cloud@pve:01234567::dGlja2V0"}}"""
    )
    with patch("requests.post", return_value=response):
        proxmox._authenticate()
    assert proxmox.csrf and proxmox.ticket
    return


def test__authenticate_failure():
    """
    Confirm that authentication failure raises an exception.
    """
    response = requests.Response()
    response.status_code = 401
    response.reason = "authentication failure"
    response.raw = io.BytesIO(b"""{"data":null}""")
    with patch("requests.post", return_value=response):
        pytest.raises(requests.exceptions.HTTPError, proxmox._authenticate)
    return


def test_creation_failure_logging(caplog):
    """
    Test detailed logging on HTTP errors during VM creation.
    """
    vm_ = {
        "profile": "my_proxmox",
        "name": "vm4",
        "technology": "lxc",
        "host": "127.0.0.1",
        "image": "local:some_image.tgz",
        "onboot": True,
    }
    assert (
        config.is_profile_configured(
            proxmox.__opts__, "my_proxmox:proxmox", "my_proxmox", vm_=vm_
        )
        is True
    )

    response = requests.Response()
    response.status_code = 400
    response.reason = "Parameter verification failed."
    response.raw = io.BytesIO(
        b"""{"data":null,"errors":{"onboot":"type check ('boolean') failed - got 'True'"}}"""
    )

    def mock_query_response(conn_type, option, post_data=None):
        if conn_type == "get" and option == "cluster/nextid":
            return 104
        if conn_type == "post" and option == "nodes/127.0.0.1/lxc":
            response.raise_for_status()
            return response
        return None

    with patch.object(proxmox, "query", side_effect=mock_query_response), patch.object(
        proxmox, "_get_properties", return_value=set()
    ):
        assert proxmox.create(vm_) is False

        # Search for these messages in a multi-line log entry.
        missing = {
            "{} Client Error: {} for url:".format(
                response.status_code, response.reason
            ),
            response.text,
        }
        for required in list(missing):
            for record in caplog.records:
                if required in record.message:
                    missing.remove(required)
                    break
        if missing:
            raise AssertionError(
                f"Did not find error messages: {sorted(list(missing))}"
            )
    return



import shutil

import pytest


@pytest.fixture(scope="module")
def state(modules):
    return modules.state


@pytest.fixture(scope="module")
def state_testfile_dest_path(tmp_path_factory):
    tempfile_dir = tmp_path_factory.mktemp("testfile-dir")
    try:
        yield tempfile_dir / "testfile"
    finally:
        shutil.rmtree(str(tempfile_dir), ignore_errors=True)


@pytest.fixture(scope="module")
def state_tree(state_tree, state_testfile_dest_path):
    top_sls_contents = """
    base:
      "*":
        - core
    """
    core_sls_contents = """
    {}:
      file:
        - managed
        - source: salt://testfile
        - makedirs: true
    """.format(
        state_testfile_dest_path
    )
    with pytest.helpers.temp_file(
        "top.sls", top_sls_contents, state_tree
    ), pytest.helpers.temp_file(
        "core.sls", core_sls_contents, state_tree
    ), pytest.helpers.temp_file(
        "testfile", "testfile base env contents", state_tree
    ):
        yield state_tree


@pytest.fixture(autouse=True)
def delete_state_testfile(state_testfile_dest_path):
    # Run Test
    try:
        yield
    finally:
        if state_testfile_dest_path.exists():
            state_testfile_dest_path.unlink()



"""
tests.pytests.integration.modules.state.test_state_state_events
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import logging
import time

import pytest

log = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def configure_state_tree(salt_master, salt_minion):
    top_file = """
    base:
      '{}':
        - state-event
    """.format(
        salt_minion.id
    )

    state_event_sls = """
    show_notification:
        test.show_notification:
            - text: Notification
    """

    with salt_master.state_tree.base.temp_file(
        "top.sls", top_file
    ), salt_master.state_tree.base.temp_file("state-event.sls", state_event_sls):
        yield


@pytest.fixture(scope="module")
def state_event_tag():
    """
    State event tag to match
    """
    return "salt/job/*/prog/{}/0"


def test_highstate_state_events(
    event_listener,
    salt_master,
    salt_minion,
    salt_call_cli,
    configure_state_tree,
    state_event_tag,
):
    """
    Test state.highstate with state_events=True
    """

    start_time = time.time()
    ret = salt_call_cli.run("state.highstate", state_events=True)
    assert ret.returncode == 0
    assert ret.data

    event_pattern = (salt_master.id, state_event_tag.format(salt_minion.id))
    matched_events = event_listener.wait_for_events(
        [event_pattern], after_time=start_time, timeout=30
    )
    assert matched_events.found_all_events


def test_sls_state_events(
    event_listener,
    salt_master,
    salt_minion,
    salt_call_cli,
    configure_state_tree,
    state_event_tag,
):
    """
    Test state.sls with state_events=True
    """

    start_time = time.time()
    ret = salt_call_cli.run("state.sls", "state-event", state_events=True)
    assert ret.returncode == 0
    assert ret.data

    event_pattern = (salt_master.id, state_event_tag.format(salt_minion.id))
    matched_events = event_listener.wait_for_events(
        [event_pattern], after_time=start_time, timeout=30
    )
    assert matched_events.found_all_events


def test_sls_id_state_events(
    event_listener,
    salt_master,
    salt_minion,
    salt_call_cli,
    configure_state_tree,
    state_event_tag,
):
    """
    Test state.sls_id with state_events=True
    """

    start_time = time.time()
    ret = salt_call_cli.run(
        "state.sls_id", "show_notification", "state-event", state_events=True
    )
    assert ret.returncode == 0
    assert ret.data

    event_pattern = (salt_master.id, state_event_tag.format(salt_minion.id))
    matched_events = event_listener.wait_for_events(
        [event_pattern], after_time=start_time, timeout=30
    )
    assert matched_events.found_all_events



import copy
import logging
import threading
import time

import pytest

from salt.utils.etcd_util import EtcdClient, EtcdClientV3, get_conn
from tests.support.pytest.etcd import *  # pylint: disable=wildcard-import,unused-wildcard-import

pytest.importorskip("docker")

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("docker", "dockerd", check_all=False),
]


@pytest.fixture(scope="module")
def minion_config_overrides(etcd_profile):
    return etcd_profile


@pytest.fixture(scope="module")
def etcd_client(minion_opts, profile_name):
    return get_conn(minion_opts, profile=profile_name)


@pytest.fixture(scope="module")
def prefix():
    return "/salt/util/test"


@pytest.fixture(autouse=True)
def cleanup_prefixed_entries(etcd_client, prefix):
    """
    Cleanup after each test to ensure a consistent starting state.
    """
    try:
        assert etcd_client.get(prefix, recurse=True) is None
        yield
    finally:
        etcd_client.delete(prefix, recurse=True)


def test_etcd_client_creation(minion_opts, profile_name, etcd_version):
    """
    Client creation using client classes, just need to assert no errors.
    """
    if etcd_version in (EtcdVersion.v2, EtcdVersion.v3_v2_mode):
        EtcdClient(minion_opts, profile=profile_name)
    else:
        EtcdClientV3(minion_opts, profile=profile_name)


def test_etcd_client_creation_with_get_conn(minion_opts, profile_name):
    """
    Client creation using get_conn, just need to assert no errors.
    """
    get_conn(minion_opts, profile=profile_name)


def test_simple_operations(etcd_client, prefix):
    """
    Verify basic functionality in order to justify use of the cleanup fixture.
    """
    assert not etcd_client.get(f"{prefix}/mtg/ambush")
    assert etcd_client.set(f"{prefix}/mtg/ambush", "viper") == "viper"
    assert etcd_client.get(f"{prefix}/mtg/ambush") == "viper"
    assert etcd_client.set(f"{prefix}/mtg/counter", "spell") == "spell"
    assert etcd_client.tree(f"{prefix}/mtg") == {
        "ambush": "viper",
        "counter": "spell",
    }
    assert etcd_client.ls(f"{prefix}/mtg") == {
        f"{prefix}/mtg": {
            f"{prefix}/mtg/ambush": "viper",
            f"{prefix}/mtg/counter": "spell",
        },
    }
    assert etcd_client.delete(f"{prefix}/mtg/ambush")
    assert etcd_client.delete(f"{prefix}/mtg", recurse=True)
    assert not etcd_client.get(f"{prefix}/mtg", recurse=True)


def test_simple_operations_with_raw_keys_and_values(
    minion_opts, profile_name, prefix, etcd_version
):
    if etcd_version in (EtcdVersion.v2, EtcdVersion.v3_v2_mode):
        pytest.skip("Not testing with raw keys using v2")
    modified_opts = copy.deepcopy(minion_opts)
    modified_opts[profile_name]["etcd.raw_keys"] = True
    modified_opts[profile_name]["etcd.raw_values"] = True
    etcd_client = get_conn(modified_opts, profile=profile_name)
    assert not etcd_client.get(f"{prefix}/mtg/ambush")
    assert etcd_client.set(f"{prefix}/mtg/ambush", "viper") == b"viper"
    assert etcd_client.get(f"{prefix}/mtg/ambush") == b"viper"
    assert etcd_client.set(f"{prefix}/mtg/counter", "spell") == b"spell"
    assert etcd_client.tree(f"{prefix}/mtg") == {
        b"ambush": b"viper",
        b"counter": b"spell",
    }
    assert etcd_client.ls(f"{prefix}/mtg") == {
        f"{prefix}/mtg".encode(): {
            f"{prefix}/mtg/ambush".encode(): b"viper",
            f"{prefix}/mtg/counter".encode(): b"spell",
        },
    }
    assert etcd_client.delete(f"{prefix}/mtg/ambush")
    assert etcd_client.delete(f"{prefix}/mtg", recurse=True)
    assert not etcd_client.get(f"{prefix}/mtg", recurse=True)


def test_get(subtests, etcd_client, prefix):
    """
    Test that get works as intended.
    """

    # Test general get case with key=value
    with subtests.test("inserted keys should be able to be retrieved"):
        etcd_client.set(f"{prefix}/get-test/key", "value")
        assert etcd_client.get(f"{prefix}/get-test/key") == "value"

    # Test with recurse=True.
    with subtests.test("keys should be able to be retrieved recursively"):
        etcd_client.set(f"{prefix}/get-test/key2/subkey", "subvalue")
        etcd_client.set(f"{prefix}/get-test/key2/subkey2/1", "subvalue1")
        etcd_client.set(f"{prefix}/get-test/key2/subkey2/2", "subvalue2")

        expected = {
            "subkey": "subvalue",
            "subkey2": {
                "1": "subvalue1",
                "2": "subvalue2",
            },
        }

        assert etcd_client.get(f"{prefix}/get-test/key2", recurse=True) == expected


def test_read(subtests, etcd_client, prefix, etcd_version):
    """
    Test that we are able to read and wait.
    """
    etcd_client.set(f"{prefix}/read/1", "one")
    etcd_client.set(f"{prefix}/read/2", "two")
    etcd_client.set(f"{prefix}/read/3/4", "three/four")

    # Simple read test
    with subtests.test(
        "reading a newly inserted and existent key should return that key"
    ):
        result = etcd_client.read(f"{prefix}/read/1")
        assert result
        if etcd_version in (EtcdVersion.v2, EtcdVersion.v3_v2_mode):
            assert result.value == "one"
        else:
            assert result.pop().value == "one"

    # Recursive read test
    with subtests.test(
        "reading recursively should return a dictionary starting at the given key"
    ):
        expected = etcd_client._flatten(
            {
                "1": "one",
                "2": "two",
                "3": {"4": "three/four"},
            },
            path=f"{prefix}/read",
        )

        result = etcd_client.read(f"{prefix}/read", recurse=True)
        assert result
        if etcd_version in (EtcdVersion.v2, EtcdVersion.v3_v2_mode):
            assert result.children
        else:
            assert len(result) > 1

        result_dict = {}
        if etcd_version in (EtcdVersion.v2, EtcdVersion.v3_v2_mode):
            for child in result.children:
                result_dict[child.key] = child.value
        else:
            for child in result:
                if child.key != f"{prefix}/read":
                    result_dict[child.key] = child.value
        assert result_dict == expected

    # Wait for an update
    with subtests.test("updates should be able to be caught by waiting in read"):
        return_list = []

        def wait_func(return_list):
            return_list.append(
                etcd_client.read(f"{prefix}/read/1", wait=True, timeout=30)
            )

        wait_thread = threading.Thread(target=wait_func, args=(return_list,))
        wait_thread.start()
        time.sleep(1)
        etcd_client.set(f"{prefix}/read/1", "not one")
        wait_thread.join()
        modified = return_list.pop()
        assert modified.key == f"{prefix}/read/1"
        assert modified.value == "not one"

    # Wait for an update using recursive
    with subtests.test("nested updates should be catchable"):
        return_list = []

        def wait_func_2(return_list):
            return_list.append(
                etcd_client.read(f"{prefix}/read", wait=True, timeout=30, recurse=True)
            )

        wait_thread = threading.Thread(target=wait_func_2, args=(return_list,))
        wait_thread.start()
        time.sleep(1)
        etcd_client.set(f"{prefix}/read/1", "one again!")
        wait_thread.join()
        modified = return_list.pop()
        assert modified.key == f"{prefix}/read/1"
        assert modified.value == "one again!"

    # Wait for an update after last modification
    with subtests.test(
        "updates should be able to be caught after an index by waiting in read"
    ):
        return_list = []
        if etcd_version in (EtcdVersion.v2, EtcdVersion.v3_v2_mode):
            last_modified = modified.modifiedIndex
        else:
            last_modified = modified.mod_revision

        def wait_func_3(return_list):
            return_list.append(
                etcd_client.read(
                    f"{prefix}/read/1",
                    wait=True,
                    timeout=30,
                    start_revision=last_modified + 1,
                )
            )

        wait_thread = threading.Thread(target=wait_func_3, args=(return_list,))
        wait_thread.start()
        time.sleep(1)
        etcd_client.set(f"{prefix}/read/1", "one")
        wait_thread.join()
        modified = return_list.pop()
        assert modified.key == f"{prefix}/read/1"
        assert modified.value == "one"

    # Wait for an update after last modification, recursively
    with subtests.test("nested updates after index should be catchable"):
        return_list = []
        if etcd_version in (EtcdVersion.v2, EtcdVersion.v3_v2_mode):
            last_modified = modified.modifiedIndex
        else:
            last_modified = modified.mod_revision

        def wait_func_4(return_list):
            return_list.append(
                etcd_client.read(
                    f"{prefix}/read",
                    wait=True,
                    timeout=30,
                    recurse=True,
                    start_revision=last_modified + 1,
                )
            )

        wait_thread = threading.Thread(target=wait_func_4, args=(return_list,))
        wait_thread.start()
        time.sleep(1)
        etcd_client.set(f"{prefix}/read/1", "one")
        wait_thread.join()
        modified = return_list.pop()
        assert modified.key == f"{prefix}/read/1"
        assert modified.value == "one"


def test_update(subtests, etcd_client, prefix):
    """
    Ensure that we can update fields
    """
    etcd_client.set(f"{prefix}/read/1", "one")
    etcd_client.set(f"{prefix}/read/2", "two")
    etcd_client.set(f"{prefix}/read/3/4", "three/four")

    # Update existent fields
    with subtests.test("update should work on already existent field"):
        updated = {
            f"{prefix}/read/1": "not one",
            f"{prefix}/read/2": "not two",
        }
        assert etcd_client.update(updated) == updated
        assert etcd_client.get(f"{prefix}/read/1") == "not one"
        assert etcd_client.get(f"{prefix}/read/2") == "not two"

    # Update non-existent fields
    with subtests.test("update should work on non-existent fields"):
        updated = {
            prefix: {
                "read-2": "read-2",
                "read-3": "read-3",
                "read-4": {
                    "sub-4": "subvalue-1",
                    "sub-4-2": "subvalue-2",
                },
            }
        }

        assert etcd_client.update(updated) == etcd_client._flatten(updated)
        assert etcd_client.get(f"{prefix}/read-2") == "read-2"
        assert etcd_client.get(f"{prefix}/read-3") == "read-3"
        assert (
            etcd_client.get(f"{prefix}/read-4", recurse=True)
            == updated[prefix]["read-4"]
        )

    with subtests.test("we should be able to prepend a path within update"):
        updated = {
            "1": "path updated one",
            "2": "path updated two",
        }
        expected_return = {
            f"{prefix}/read/1": "path updated one",
            f"{prefix}/read/2": "path updated two",
        }
        assert etcd_client.update(updated, path=f"{prefix}/read") == expected_return
        assert etcd_client.get(f"{prefix}/read/1") == "path updated one"
        assert etcd_client.get(f"{prefix}/read/2") == "path updated two"


def test_write_file(subtests, etcd_client, prefix):
    """
    Test solely writing files
    """
    with subtests.test(
        "we should be able to write a single value for a non-existent key"
    ):
        assert etcd_client.write_file(f"{prefix}/write/key_1", "value_1") == "value_1"
        assert etcd_client.get(f"{prefix}/write/key_1") == "value_1"

    with subtests.test("we should be able to write a single value for an existent key"):
        assert (
            etcd_client.write_file(f"{prefix}/write/key_1", "new_value_1")
            == "new_value_1"
        )
        assert etcd_client.get(f"{prefix}/write/key_1") == "new_value_1"

    with subtests.test("we should be able to write a single value with a ttl"):
        assert (
            etcd_client.write_file(f"{prefix}/write/ttl_key", "new_value_2", ttl=5)
            == "new_value_2"
        )
        time.sleep(10)
        assert etcd_client.get(f"{prefix}/write/ttl_key") is None


def test_write_directory(subtests, etcd_client, prefix, etcd_version):
    """
    Test solely writing directories
    """
    if etcd_version != EtcdVersion.v2:
        pytest.skip("write_directory is not defined for etcd v3")

    with subtests.test("we should be able to create a non-existent directory"):
        assert etcd_client.write_directory(f"{prefix}/write_dir/dir1", None)
        assert etcd_client.get(f"{prefix}/write_dir/dir1") is None

    with subtests.test("writing an already existent directory should return True"):
        assert etcd_client.write_directory(f"{prefix}/write_dir/dir1", None)
        assert etcd_client.get(f"{prefix}/write_dir/dir1") is None

    with subtests.test("we should be able to write to a new directory"):
        assert (
            etcd_client.write_file(f"{prefix}/write_dir/dir1/key1", "value1")
            == "value1"
        )
        assert etcd_client.get(f"{prefix}/write_dir/dir1/key1") == "value1"


def test_ls(subtests, etcd_client, prefix):
    """
    Test listing top level contents
    """
    with subtests.test("ls on a non-existent directory should return an empty dict"):
        assert not etcd_client.ls(f"{prefix}/ls")

    with subtests.test(
        "ls should list the top level keys and values at the given path"
    ):
        etcd_client.set(f"{prefix}/ls/1", "one")
        etcd_client.set(f"{prefix}/ls/2", "two")
        etcd_client.set(f"{prefix}/ls/3/4", "three/four")

        # If it's a dir, it's suffixed with a slash
        expected = {
            f"{prefix}/ls": {
                f"{prefix}/ls/1": "one",
                f"{prefix}/ls/2": "two",
                f"{prefix}/ls/3/": {},
            },
        }

        assert etcd_client.ls(f"{prefix}/ls") == expected


@pytest.mark.parametrize("func", ("rm", "delete"))
def test_rm_and_delete(subtests, etcd_client, prefix, func, etcd_version):
    """
    Ensure we can remove keys using rm
    """
    func = getattr(etcd_client, func)

    with subtests.test("removing a non-existent key should do nothing"):
        assert func(f"{prefix}/rm/key1") is None

    with subtests.test("we should be able to remove an existing key"):
        etcd_client.set(f"{prefix}/rm/key1", "value1")
        assert func(f"{prefix}/rm/key1")
        assert etcd_client.get(f"{prefix}/rm/key1") is None

    with subtests.test("we should be able to remove an empty directory"):
        if etcd_version == EtcdVersion.v2:
            etcd_client.write_directory(f"{prefix}/rm/dir1", None)
            assert func(f"{prefix}/rm/dir1", recurse=True)
            assert etcd_client.get(f"{prefix}/rm/dir1", recurse=True) is None

    with subtests.test("we should be able to remove a directory with keys"):
        updated = {
            "dir1": {
                "rm-1": "value-1",
                "rm-2": {
                    "sub-rm-1": "subvalue-1",
                    "sub-rm-2": "subvalue-2",
                },
            }
        }
        etcd_client.update(updated, path=f"{prefix}/rm")

        assert func(f"{prefix}/rm/dir1", recurse=True)
        assert etcd_client.get(f"{prefix}/rm/dir1", recurse=True) is None
        assert etcd_client.get(f"{prefix}/rm/dir1/rm-1", recurse=True) is None

    with subtests.test("removing a directory without recursion should do nothing"):
        updated = {
            "dir1": {
                "rm-1": "value-1",
                "rm-2": {
                    "sub-rm-1": "subvalue-1",
                    "sub-rm-2": "subvalue-2",
                },
            }
        }
        etcd_client.update(updated, path=f"{prefix}/rm")

        assert func(f"{prefix}/rm/dir1") is None
        assert etcd_client.get(f"{prefix}/rm/dir1", recurse=True) == updated["dir1"]
        assert etcd_client.get(f"{prefix}/rm/dir1/rm-1") == "value-1"


def test_tree(subtests, etcd_client, prefix, etcd_version):
    """
    Tree should return a dictionary representing what is downstream of the prefix.
    """
    with subtests.test("the tree of a non-existent key should be None"):
        assert etcd_client.tree(prefix) is None

    with subtests.test("the tree of an file should be {key: value}"):
        etcd_client.set(f"{prefix}/1", "one")
        assert etcd_client.tree(f"{prefix}/1") == {"1": "one"}

    with subtests.test("the tree of an empty directory should be empty"):
        if etcd_version == EtcdVersion.v2:
            etcd_client.write_directory(f"{prefix}/2", None)
            assert etcd_client.tree(f"{prefix}/2") == {}

    with subtests.test("we should be able to recieve the tree of a directory"):
        etcd_client.set(f"{prefix}/3/4", "three/four")
        expected = {
            "1": "one",
            "2": {},
            "3": {"4": "three/four"},
        }
        if etcd_version != EtcdVersion.v2:
            expected.pop("2")
        assert etcd_client.tree(prefix) == expected

    with subtests.test("we should be able to recieve the tree of an outer directory"):
        etcd_client.set(f"{prefix}/5/6/7", "five/six/seven")
        expected = {
            "6": {"7": "five/six/seven"},
        }
        assert etcd_client.tree(f"{prefix}/5") == expected


def test_watch(subtests, etcd_client, prefix):
    updated = {
        "1": "one",
        "2": "two",
        "3": {
            "4": "three/four",
        },
    }
    etcd_client.update(updated, path=f"{prefix}/watch")

    with subtests.test("watching an invalid key should timeout and return None"):
        assert etcd_client.watch(f"{prefix}/invalid", timeout=3) is None

    with subtests.test(
        "watching an valid key with no changes should timeout and return None"
    ):
        assert etcd_client.watch(f"{prefix}/watch/1", timeout=3) is None

    # Wait for an update
    with subtests.test("updates should be able to be caught by waiting in read"):
        return_list = []

        def wait_func(return_list):
            return_list.append(etcd_client.watch(f"{prefix}/watch/1", timeout=30))

        wait_thread = threading.Thread(target=wait_func, args=(return_list,))
        wait_thread.start()
        time.sleep(1)
        etcd_client.set(f"{prefix}/watch/1", "not one")
        wait_thread.join()
        modified = return_list.pop()
        assert modified["key"] == f"{prefix}/watch/1"
        assert modified["value"] == "not one"

    # Wait for an update using recursive
    with subtests.test("nested updates should be catchable"):
        return_list = []

        def wait_func_2(return_list):
            return_list.append(
                etcd_client.watch(f"{prefix}/watch", timeout=30, recurse=True)
            )

        wait_thread = threading.Thread(target=wait_func_2, args=(return_list,))
        wait_thread.start()
        time.sleep(1)
        etcd_client.set(f"{prefix}/watch/1", "one again!")
        wait_thread.join()
        modified = return_list.pop()
        assert modified["key"] == f"{prefix}/watch/1"
        assert modified["value"] == "one again!"

    # Wait for an update after last modification
    with subtests.test(
        "updates should be able to be caught after an index by waiting in read"
    ):
        return_list = []
        last_modified = modified["mIndex"]

        def wait_func_3(return_list):
            return_list.append(
                etcd_client.watch(
                    f"{prefix}/watch/1",
                    timeout=30,
                    start_revision=last_modified + 1,
                )
            )

        wait_thread = threading.Thread(target=wait_func_3, args=(return_list,))
        wait_thread.start()
        time.sleep(1)
        etcd_client.set(f"{prefix}/watch/1", "one")
        wait_thread.join()
        modified = return_list.pop()
        assert modified["key"] == f"{prefix}/watch/1"
        assert modified["value"] == "one"

    # Wait for an update after last modification, recursively
    with subtests.test("nested updates after index should be catchable"):
        return_list = []
        last_modified = modified["mIndex"]

        def wait_func_4(return_list):
            return_list.append(
                etcd_client.watch(
                    f"{prefix}/watch",
                    timeout=30,
                    recurse=True,
                    start_revision=last_modified + 1,
                )
            )

        wait_thread = threading.Thread(target=wait_func_4, args=(return_list,))
        wait_thread.start()
        time.sleep(1)
        etcd_client.set(f"{prefix}/watch/1", "one")
        wait_thread.join()
        modified = return_list.pop()
        assert modified["key"] == f"{prefix}/watch/1"
        assert modified["value"] == "one"



"""
Control the state system on the minion.

State Caching
-------------

When a highstate is called, the minion automatically caches a copy of the last
high data. If you then run a highstate with cache=True it will use that cached
highdata and won't hit the fileserver except for ``salt://`` links in the
states themselves.
"""

import logging
import os
import shutil
import sys
import tarfile
import tempfile
import time
from collections import OrderedDict

import salt.config
import salt.defaults.exitcodes
import salt.payload
import salt.state
import salt.utils.args
import salt.utils.data
import salt.utils.event
import salt.utils.files
import salt.utils.functools
import salt.utils.hashutils
import salt.utils.jid
import salt.utils.json
import salt.utils.msgpack
import salt.utils.platform
import salt.utils.state
import salt.utils.stringutils
import salt.utils.url
import salt.utils.versions
from salt.exceptions import CommandExecutionError, SaltInvocationError
from salt.loader import _format_cached_grains
from salt.runners.state import orchestrate as _orchestrate

__proxyenabled__ = ["*"]

__outputter__ = {
    "sls": "highstate",
    "sls_id": "highstate",
    "pkg": "highstate",
    "top": "highstate",
    "single": "highstate",
    "highstate": "highstate",
    "template": "highstate",
    "template_str": "highstate",
    "apply_": "highstate",
    "test": "highstate",
    "request": "highstate",
    "check_request": "highstate",
    "run_request": "highstate",
}

__func_alias__ = {"apply_": "apply"}
log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "state"


def __virtual__():
    """
    Set the virtualname
    """
    # Update global namespace with functions that are cloned in this module
    global _orchestrate
    _orchestrate = salt.utils.functools.namespaced_function(_orchestrate, globals())

    return __virtualname__


def _filter_running(runnings):
    """
    Filter out the result: True + no changes data
    """
    ret = {
        tag: value
        for tag, value in runnings.items()
        if not value["result"] or value["changes"]
    }
    return ret


def _set_retcode(ret, highstate=None):
    """
    Set the return code based on the data back from the state system
    """

    # Set default retcode to 0
    __context__["retcode"] = salt.defaults.exitcodes.EX_OK

    if isinstance(ret, list):
        __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
        return
    if not __utils__["state.check_result"](ret, highstate=highstate):
        __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_FAILURE


def _get_pillar_errors(kwargs, pillar=None):
    """
    Check pillar for errors.

    If a pillar is passed, it will be checked. Otherwise, the in-memory pillar
    will checked instead. Passing kwargs['force'] = True short cuts the check
    and always returns None, indicating no errors.

    :param kwargs: dictionary of options
    :param pillar: pillar
    :return: None or a list of error messages
    """
    return None if kwargs.get("force") else (pillar or __pillar__).get("_errors")


def _wait(jid, max_queue=0):
    """
    Wait for all previously started state jobs to finish running
    """
    if jid is None:
        jid = salt.utils.jid.gen_jid(__opts__)
    states = _prior_running_states(jid)
    if not max_queue or len(states) < max_queue:
        while states:
            time.sleep(1)
            states = _prior_running_states(jid)
        return True
    return False


def _snapper_pre(opts, jid):
    """
    Create a snapper pre snapshot
    """
    snapper_pre = None
    try:
        if not opts["test"] and __opts__.get("snapper_states"):
            # Run the snapper pre snapshot
            snapper_pre = __salt__["snapper.create_snapshot"](
                config=__opts__.get("snapper_states_config", "root"),
                snapshot_type="pre",
                description=f"Salt State run for jid {jid}",
                __pub_jid=jid,
            )
    except Exception:  # pylint: disable=broad-except
        log.error("Failed to create snapper pre snapshot for jid: %s", jid)
    return snapper_pre


def _snapper_post(opts, jid, pre_num):
    """
    Create the post states snapshot
    """
    try:
        if not opts["test"] and __opts__.get("snapper_states") and pre_num:
            # Run the snapper pre snapshot
            __salt__["snapper.create_snapshot"](
                config=__opts__.get("snapper_states_config", "root"),
                snapshot_type="post",
                pre_number=pre_num,
                description=f"Salt State run for jid {jid}",
                __pub_jid=jid,
            )
    except Exception:  # pylint: disable=broad-except
        log.error("Failed to create snapper pre snapshot for jid: %s", jid)


def _get_pause(jid, state_id=None):
    """
    Return the pause information for a given jid
    """
    pause_dir = os.path.join(__opts__["cachedir"], "state_pause")
    pause_path = os.path.join(pause_dir, jid)
    if not os.path.exists(pause_dir):
        try:
            os.makedirs(pause_dir)
        except OSError:
            # File created in the gap
            pass
    data = {}
    if state_id is not None:
        if state_id not in data:
            data[state_id] = {}
    if os.path.exists(pause_path):
        with salt.utils.files.fopen(pause_path, "rb") as fp_:
            data = salt.utils.msgpack.loads(fp_.read())
    return data, pause_path


def get_pauses(jid=None):
    """
    Get a report on all of the currently paused state runs and pause
    run settings.
    Optionally send in a jid if you only desire to see a single pause
    data set.
    """
    ret = {}
    active = __salt__["saltutil.is_running"]("state.*")
    pause_dir = os.path.join(__opts__["cachedir"], "state_pause")
    if not os.path.exists(pause_dir):
        return ret
    if jid is None:
        jids = os.listdir(pause_dir)
    elif isinstance(jid, list):
        jids = salt.utils.data.stringify(jid)
    else:
        jids = [str(jid)]
    for scan_jid in jids:
        is_active = False
        for active_data in active:
            if active_data["jid"] == scan_jid:
                is_active = True
        if not is_active:
            try:
                pause_path = os.path.join(pause_dir, scan_jid)
                os.remove(pause_path)
            except OSError:
                # Already gone
                pass
            continue
        data, pause_path = _get_pause(scan_jid)
        ret[scan_jid] = data
    return ret


def soft_kill(jid, state_id=None):
    """
    Set up a state run to die before executing the given state id,
    this instructs a running state to safely exit at a given
    state id. This needs to pass in the jid of the running state.
    If a state_id is not passed then the jid referenced will be safely exited
    at the beginning of the next state run.

    The given state id is the id got a given state execution, so given a state
    that looks like this:

    .. code-block:: yaml

        vim:
          pkg.installed: []

    The state_id to pass to `soft_kill` is `vim`

    CLI Examples:

    .. code-block:: bash

        salt '*' state.soft_kill 20171130110407769519
        salt '*' state.soft_kill 20171130110407769519 vim
    """
    jid = str(jid)
    if state_id is None:
        state_id = "__all__"
    data, pause_path = _get_pause(jid, state_id)
    data[state_id]["kill"] = True
    with salt.utils.files.fopen(pause_path, "wb") as fp_:
        fp_.write(salt.utils.msgpack.dumps(data))


def pause(jid, state_id=None, duration=None):
    """
    Set up a state id pause, this instructs a running state to pause at a given
    state id. This needs to pass in the jid of the running state and can
    optionally pass in a duration in seconds. If a state_id is not passed then
    the jid referenced will be paused at the beginning of the next state run.

    The given state id is the id got a given state execution, so given a state
    that looks like this:

    .. code-block:: yaml

        vim:
          pkg.installed: []

    The state_id to pass to `pause` is `vim`

    CLI Examples:

    .. code-block:: bash

        salt '*' state.pause 20171130110407769519
        salt '*' state.pause 20171130110407769519 vim
        salt '*' state.pause 20171130110407769519 vim 20
    """
    jid = str(jid)
    if state_id is None:
        state_id = "__all__"
    data, pause_path = _get_pause(jid, state_id)
    if duration:
        data[state_id]["duration"] = int(duration)
    with salt.utils.files.fopen(pause_path, "wb") as fp_:
        fp_.write(salt.utils.msgpack.dumps(data))


def resume(jid, state_id=None):
    """
    Remove a pause from a jid, allowing it to continue. If the state_id is
    not specified then the a general pause will be resumed.

    The given state_id is the id got a given state execution, so given a state
    that looks like this:

    .. code-block:: yaml

        vim:
          pkg.installed: []

    The state_id to pass to `rm_pause` is `vim`

    CLI Examples:

    .. code-block:: bash

        salt '*' state.resume 20171130110407769519
        salt '*' state.resume 20171130110407769519 vim
    """
    jid = str(jid)
    if state_id is None:
        state_id = "__all__"
    data, pause_path = _get_pause(jid, state_id)
    if state_id in data:
        data.pop(state_id)
    if state_id == "__all__":
        data = {}
    with salt.utils.files.fopen(pause_path, "wb") as fp_:
        fp_.write(salt.utils.msgpack.dumps(data))


def orchestrate(
    mods, saltenv="base", test=None, exclude=None, pillar=None, pillarenv=None
):
    """
    .. versionadded:: 2016.11.0

    Execute the orchestrate runner from a masterless minion.

    .. seealso:: More Orchestrate documentation

        * :ref:`Full Orchestrate Tutorial <orchestrate-runner>`
        * Docs for the salt state module :py:mod:`salt.states.saltmod`

    CLI Examples:

    .. code-block:: bash

        salt-call --local state.orchestrate webserver
        salt-call --local state.orchestrate webserver saltenv=dev test=True
        salt-call --local state.orchestrate webserver saltenv=dev pillarenv=aws
    """
    return _orchestrate(
        mods=mods,
        saltenv=saltenv,
        test=test,
        exclude=exclude,
        pillar=pillar,
        pillarenv=pillarenv,
    )


def running(concurrent=False):
    """
    Return a list of strings that contain state return data if a state function
    is already running. This function is used to prevent multiple state calls
    from being run at the same time.

    CLI Example:

    .. code-block:: bash

        salt '*' state.running
    """
    ret = []
    if concurrent:
        return ret
    active = __salt__["saltutil.is_running"]("state.*")
    for data in active:
        err = (
            'The function "{}" is running as PID {} and was started at {} '
            "with jid {}".format(
                data["fun"],
                data["pid"],
                salt.utils.jid.jid_to_time(data["jid"]),
                data["jid"],
            )
        )
        ret.append(err)
    return ret


def _prior_running_states(jid):
    """
    Return a list of dicts of prior calls to state functions.  This function is
    used to queue state calls so only one is run at a time.
    """

    ret = []
    active = __salt__["saltutil.is_running"]("state.*")
    for data in active:
        try:
            data_jid = int(data["jid"])
        except ValueError:
            continue
        if data_jid < int(jid):
            ret.append(data)
    return ret


def _check_queue(queue, kwargs):
    """
    Utility function to queue the state run if requested
    and to check for conflicts in currently running states
    """
    if queue is None:
        queue = __salt__["config.option"]("state_queue", False)

    if queue is True:
        _wait(kwargs.get("__pub_jid"))
    else:
        queue_ret = False
        if not isinstance(queue, bool) and isinstance(queue, int):
            queue_ret = _wait(kwargs.get("__pub_jid"), max_queue=queue)

        if not queue_ret:
            conflict = running(concurrent=kwargs.get("concurrent", False))
            if conflict:
                __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
                return conflict

    return


def _get_initial_pillar(opts):
    return (
        __pillar__.value()
        if __opts__.get("__cli", None) == "salt-call"
        and opts["pillarenv"] == __opts__["pillarenv"]
        else None
    )


def low(data, queue=None, **kwargs):
    """
    Execute a single low data call

    This function is mostly intended for testing the state system and is not
    likely to be needed in everyday usage.

    CLI Example:

    .. code-block:: bash

        salt '*' state.low '{"state": "pkg", "fun": "installed", "name": "vi"}'
    """
    conflict = _check_queue(queue, kwargs)
    if conflict is not None:
        return conflict
    try:
        st_ = salt.state.State(__opts__, proxy=__proxy__)
    except NameError:
        st_ = salt.state.State(__opts__)
    err = st_.verify_data(data)
    if err:
        __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
        return err
    ret = st_.call(data)
    if isinstance(ret, list):
        __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
    if __utils__["state.check_result"](ret):
        __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_FAILURE
    return ret


def _get_test_value(test=None, **kwargs):
    """
    Determine the correct value for the test flag.
    """
    ret = True
    if test is None:
        if salt.utils.args.test_mode(test=test, **kwargs):
            ret = True
        elif __salt__["config.get"]("test", omit_opts=True) is True:
            ret = True
        else:
            ret = __opts__.get("test", None)
    elif test is False:
        ret = False
    return ret


def high(data, test=None, queue=None, **kwargs):
    """
    Execute the compound calls stored in a single set of high data

    This function is mostly intended for testing the state system and is not
    likely to be needed in everyday usage.

    CLI Example:

    .. code-block:: bash

        salt '*' state.high '{"vim": {"pkg": ["installed"]}}'
    """
    conflict = _check_queue(queue, kwargs)
    if conflict is not None:
        return conflict
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)

    opts["test"] = _get_test_value(test, **kwargs)

    pillar_override = kwargs.get("pillar")
    pillar_enc = kwargs.get("pillar_enc")
    if (
        pillar_enc is None
        and pillar_override is not None
        and not isinstance(pillar_override, dict)
    ):
        raise SaltInvocationError(
            "Pillar data must be formatted as a dictionary, unless pillar_enc "
            "is specified."
        )
    try:
        st_ = salt.state.State(
            opts,
            pillar_override,
            pillar_enc=pillar_enc,
            proxy=dict(__proxy__),
            context=dict(__context__),
            initial_pillar=_get_initial_pillar(opts),
        )
    except NameError:
        st_ = salt.state.State(
            opts,
            pillar_override,
            pillar_enc=pillar_enc,
            initial_pillar=_get_initial_pillar(opts),
        )

    ret = st_.call_high(data)
    _set_retcode(ret, highstate=data)
    return ret


def template(tem, queue=None, **kwargs):
    """
    Execute the information stored in a template file on the minion.

    This function does not ask a master for a SLS file to render but
    instead directly processes the file at the provided path on the minion.

    CLI Example:

    .. code-block:: bash

        salt '*' state.template '<Path to template on the minion>'
    """
    if "env" in kwargs:
        # "env" is not supported; Use "saltenv".
        kwargs.pop("env")

    conflict = _check_queue(queue, kwargs)
    if conflict is not None:
        return conflict

    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    try:
        st_ = salt.state.HighState(
            opts,
            context=dict(__context__),
            proxy=dict(__proxy__),
            initial_pillar=_get_initial_pillar(opts),
        )
    except NameError:
        st_ = salt.state.HighState(
            opts, context=dict(__context__), initial_pillar=_get_initial_pillar(opts)
        )

    with st_:
        errors = _get_pillar_errors(kwargs, pillar=st_.opts["pillar"])
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            raise CommandExecutionError("Pillar failed to render", info=errors)

        if not tem.endswith(".sls"):
            tem = f"{tem}.sls"
        high_state, errors = st_.render_state(
            tem, kwargs.get("saltenv", ""), "", None, local=True
        )
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
            return errors
        ret = st_.state.call_high(high_state)
        _set_retcode(ret, highstate=high_state)
        return ret


def template_str(tem, queue=None, **kwargs):
    """
    Execute the information stored in a string from an sls template

    CLI Example:

    .. code-block:: bash

        salt '*' state.template_str '<Template String>'
    """
    conflict = _check_queue(queue, kwargs)
    if conflict is not None:
        return conflict

    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)

    try:
        st_ = salt.state.State(
            opts, proxy=__proxy__, initial_pillar=_get_initial_pillar(opts)
        )
    except NameError:
        st_ = salt.state.State(opts, initial_pillar=_get_initial_pillar(opts))
    ret = st_.call_template_str(tem)
    _set_retcode(ret)
    return ret


def apply_(mods=None, **kwargs):
    """
    .. versionadded:: 2015.5.0

    This function will call :mod:`state.highstate
    <salt.modules.state.highstate>` or :mod:`state.sls
    <salt.modules.state.sls>` based on the arguments passed to this function.
    It exists as a more intuitive way of applying states.

    .. rubric:: APPLYING ALL STATES CONFIGURED IN TOP.SLS (A.K.A. :ref:`HIGHSTATE <running-highstate>`)

    To apply all configured states, simply run ``state.apply`` with no SLS
    targets, like so:

    .. code-block:: bash

        salt '*' state.apply

    The following additional arguments are also accepted when applying all
    states configured in top.sls:

    test
        Run states in test-only (dry-run) mode

    mock
        The mock option allows for the state run to execute without actually
        calling any states. This then returns a mocked return which will show
        the requisite ordering as well as fully validate the state run.

        .. versionadded:: 2015.8.4

    pillar
        Custom Pillar values, passed as a dictionary of key-value pairs

        .. code-block:: bash

            salt '*' state.apply stuff pillar='{"foo": "bar"}'

        .. note::
            Values passed this way will override Pillar values set via
            ``pillar_roots`` or an external Pillar source.

    exclude
        Exclude specific states from execution. Accepts a list of sls names, a
        comma-separated string of sls names, or a list of dictionaries
        containing ``sls`` or ``id`` keys. Glob-patterns may be used to match
        multiple states.

        .. code-block:: bash

            salt '*' state.apply exclude=bar,baz
            salt '*' state.apply exclude=foo*
            salt '*' state.apply exclude="[{'id': 'id_to_exclude'}, {'sls': 'sls_to_exclude'}]"

    queue : False
        Instead of failing immediately when another state run is in progress,
        a value of ``True`` will queue the new state run to begin running once
        the other has finished.

        This option starts a new thread for each queued state run, so use this
        option sparingly.

        .. versionchanged:: 3006.0
            This parameter can also be set via the ``state_queue`` configuration
            option. Additionally, it can now be set to an integer representing
            the maximum queue size which can be attained before the state runs
            will fail to be queued. This can prevent runaway conditions where
            new threads are started until system performance is hampered.

    localconfig
        Optionally, instead of using the minion config, load minion opts from
        the file specified by this argument, and then merge them with the
        options from the minion config. This functionality allows for specific
        states to be run with their own custom minion configuration, including
        different pillars, file_roots, etc.

        .. code-block:: bash

            salt '*' state.apply localconfig=/path/to/minion.yml

    state_events
        The state_events option sends progress events as each function in
        a state run completes execution.

        .. versionadded:: 3006.0


    .. rubric:: APPLYING INDIVIDUAL SLS FILES (A.K.A. :py:func:`STATE.SLS <salt.modules.state.sls>`)

    To apply individual SLS files, pass them as a comma-separated list:

    .. code-block:: bash

        # Run the states configured in salt://stuff.sls (or salt://stuff/init.sls)
        salt '*' state.apply stuff

        # Run the states configured in salt://stuff.sls (or salt://stuff/init.sls)
        # and salt://pkgs.sls (or salt://pkgs/init.sls).
        salt '*' state.apply stuff,pkgs

        # Run the states configured in a more deeply nested directory such as salt://my/organized/stuff.sls (or salt://my/organized/stuff/init.sls)
        salt '*' state.apply my.organized.stuff

    The following additional arguments are also accepted when applying
    individual SLS files:

    test
        Run states in test-only (dry-run) mode

    mock
        The mock option allows for the state run to execute without actually
        calling any states. This then returns a mocked return which will show
        the requisite ordering as well as fully validate the state run.

        .. versionadded:: 2015.8.4

    pillar
        Custom Pillar values, passed as a dictionary of key-value pairs

        .. code-block:: bash

            salt '*' state.apply stuff pillar='{"foo": "bar"}'

        .. note::
            Values passed this way will override Pillar values set via
            ``pillar_roots`` or an external Pillar source.

    queue : False
        Instead of failing immediately when another state run is in progress,
        a value of ``True`` will queue the new state run to begin running once
        the other has finished.

        This option starts a new thread for each queued state run, so use this
        option sparingly.

        .. versionchanged:: 3006.0
            This parameter can also be set via the ``state_queue`` configuration
            option. Additionally, it can now be set to an integer representing
            the maximum queue size which can be attained before the state runs
            will fail to be queued. This can prevent runaway conditions where
            new threads are started until system performance is hampered.

    concurrent : False
        Execute state runs concurrently instead of serially

        .. warning::

            This flag is potentially dangerous. It is designed for use when
            multiple state runs can safely be run at the same time. Do *not*
            use this flag for performance optimization.

    saltenv
        Specify a salt fileserver environment to be used when applying states

        .. versionchanged:: 0.17.0
            Argument name changed from ``env`` to ``saltenv``

        .. versionchanged:: 2014.7.0
            If no saltenv is specified, the minion config will be checked for an
            ``environment`` parameter and if found, it will be used. If none is
            found, ``base`` will be used. In prior releases, the minion config
            was not checked and ``base`` would always be assumed when the
            saltenv was not explicitly set.

    pillarenv
        Specify a Pillar environment to be used when applying states. This
        can also be set in the minion config file using the
        :conf_minion:`pillarenv` option. When neither the
        :conf_minion:`pillarenv` minion config option nor this CLI argument is
        used, all Pillar environments will be merged together.

    localconfig
        Optionally, instead of using the minion config, load minion opts from
        the file specified by this argument, and then merge them with the
        options from the minion config. This functionality allows for specific
        states to be run with their own custom minion configuration, including
        different pillars, file_roots, etc.

        .. code-block:: bash

            salt '*' state.apply stuff localconfig=/path/to/minion.yml

    sync_mods
        If specified, the desired custom module types will be synced prior to
        running the SLS files:

        .. code-block:: bash

            salt '*' state.apply stuff sync_mods=states,modules
            salt '*' state.apply stuff sync_mods=all

        .. note::
            This option is ignored when no SLS files are specified, as a
            :ref:`highstate <running-highstate>` automatically syncs all custom
            module types.

        .. versionadded:: 2017.7.8,2018.3.3,2019.2.0

    state_events
        The state_events option sends progress events as each function in
        a state run completes execution.

        .. versionadded:: 3006.0
    """
    if mods:
        return sls(mods, **kwargs)
    return highstate(**kwargs)


def test(*args, **kwargs):
    """
    .. versionadded:: 3001

    Alias for `state.apply` with the kwarg `test` forced to `True`.

    This is a nicety to avoid the need to type out `test=True` and the possibility of
    a typo causing changes you do not intend.
    """
    kwargs["test"] = True
    ret = apply_(*args, **kwargs)

    return ret


def request(mods=None, **kwargs):
    """
    .. versionadded:: 2015.5.0

    Request that the local admin execute a state run via
    `salt-call state.run_request`.
    All arguments match those of state.apply.

    CLI Example:

    .. code-block:: bash

        salt '*' state.request
        salt '*' state.request stuff
        salt '*' state.request stuff,pkgs
    """
    kwargs["test"] = True
    ret = apply_(mods, **kwargs)
    notify_path = os.path.join(__opts__["cachedir"], "req_state.p")
    req = check_request()
    req.update(
        {
            kwargs.get("name", "default"): {
                "test_run": ret,
                "mods": mods,
                "kwargs": kwargs,
            }
        }
    )
    with salt.utils.files.set_umask(0o077):
        try:
            if salt.utils.platform.is_windows():
                # Make sure cache file isn't read-only
                __salt__["cmd.run"](f'attrib -R "{notify_path}"')
            with salt.utils.files.fopen(notify_path, "w+b") as fp_:
                salt.payload.dump(req, fp_)
        except OSError:
            log.error(
                "Unable to write state request file %s. Check permission.", notify_path
            )
    return ret


def check_request(name=None):
    """
    .. versionadded:: 2015.5.0

    Return the state request information, if any

    CLI Example:

    .. code-block:: bash

        salt '*' state.check_request
    """
    notify_path = os.path.join(__opts__["cachedir"], "req_state.p")
    if os.path.isfile(notify_path):
        with salt.utils.files.fopen(notify_path, "rb") as fp_:
            req = salt.payload.load(fp_)
        if name:
            return req[name]
        return req
    return {}


def clear_request(name=None):
    """
    .. versionadded:: 2015.5.0

    Clear out the state execution request without executing it

    CLI Example:

    .. code-block:: bash

        salt '*' state.clear_request
    """
    notify_path = os.path.join(__opts__["cachedir"], "req_state.p")
    if not os.path.isfile(notify_path):
        return True
    if not name:
        try:
            os.remove(notify_path)
        except OSError:
            pass
    else:
        req = check_request()
        if name in req:
            req.pop(name)
        else:
            return False
        with salt.utils.files.set_umask(0o077):
            try:
                if salt.utils.platform.is_windows():
                    # Make sure cache file isn't read-only
                    __salt__["cmd.run"](f'attrib -R "{notify_path}"')
                with salt.utils.files.fopen(notify_path, "w+b") as fp_:
                    salt.payload.dump(req, fp_)
            except OSError:
                log.error(
                    "Unable to write state request file %s. Check permission.",
                    notify_path,
                )
    return True


def run_request(name="default", **kwargs):
    """
    .. versionadded:: 2015.5.0

    Execute the pending state request

    CLI Example:

    .. code-block:: bash

        salt '*' state.run_request
    """
    req = check_request()
    if name not in req:
        return {}
    n_req = req[name]
    if "mods" not in n_req or "kwargs" not in n_req:
        return {}
    req[name]["kwargs"].update(kwargs)
    if "test" in n_req["kwargs"]:
        n_req["kwargs"].pop("test")
    if req:
        ret = apply_(n_req["mods"], **n_req["kwargs"])
        try:
            os.remove(os.path.join(__opts__["cachedir"], "req_state.p"))
        except OSError:
            pass
        return ret
    return {}


def highstate(test=None, queue=None, state_events=None, **kwargs):
    """
    Retrieve the state data from the salt master for this minion and execute it

    test
        Run states in test-only (dry-run) mode

    pillar
        Custom Pillar values, passed as a dictionary of key-value pairs

        .. code-block:: bash

            salt '*' state.highstate stuff pillar='{"foo": "bar"}'

        .. note::
            Values passed this way will override Pillar values set via
            ``pillar_roots`` or an external Pillar source.

        .. versionchanged:: 2016.3.0
            GPG-encrypted CLI Pillar data is now supported via the GPG
            renderer. See :ref:`here <encrypted-cli-pillar-data>` for details.

    pillar_enc
        Specify which renderer to use to decrypt encrypted data located within
        the ``pillar`` value. Currently, only ``gpg`` is supported.

        .. versionadded:: 2016.3.0

    exclude
        Exclude specific states from execution. Accepts a list of sls names, a
        comma-separated string of sls names, or a list of dictionaries
        containing ``sls`` or ``id`` keys. Glob-patterns may be used to match
        multiple states.

        .. code-block:: bash

            salt '*' state.highstate exclude=bar,baz
            salt '*' state.highstate exclude=foo*
            salt '*' state.highstate exclude="[{'id': 'id_to_exclude'}, {'sls': 'sls_to_exclude'}]"

    saltenv
        Specify a salt fileserver environment to be used when applying states

        .. versionchanged:: 0.17.0
            Argument name changed from ``env`` to ``saltenv``.

        .. versionchanged:: 2014.7.0
            If no saltenv is specified, the minion config will be checked for a
            ``saltenv`` parameter and if found, it will be used. If none is
            found, ``base`` will be used. In prior releases, the minion config
            was not checked and ``base`` would always be assumed when the
            saltenv was not explicitly set.

    pillarenv
        Specify a Pillar environment to be used when applying states. This
        can also be set in the minion config file using the
        :conf_minion:`pillarenv` option. When neither the
        :conf_minion:`pillarenv` minion config option nor this CLI argument is
        used, all Pillar environments will be merged together.

    queue : False
        Instead of failing immediately when another state run is in progress,
        a value of ``True`` will queue the new state run to begin running once
        the other has finished.

        This option starts a new thread for each queued state run, so use this
        option sparingly.

        .. versionchanged:: 3006.0
            This parameter can also be set via the ``state_queue`` configuration
            option. Additionally, it can now be set to an integer representing
            the maximum queue size which can be attained before the state runs
            will fail to be queued. This can prevent runaway conditions where
            new threads are started until system performance is hampered.

    concurrent : False
        Execute state runs concurrently instead of serially

        .. warning::

            This flag is potentially dangerous. It is designed for use when
            multiple state runs can safely be run at the same time. Do *not*
            use this flag for performance optimization.

    localconfig
        Optionally, instead of using the minion config, load minion opts from
        the file specified by this argument, and then merge them with the
        options from the minion config. This functionality allows for specific
        states to be run with their own custom minion configuration, including
        different pillars, file_roots, etc.

    mock
        The mock option allows for the state run to execute without actually
        calling any states. This then returns a mocked return which will show
        the requisite ordering as well as fully validate the state run.

        .. versionadded:: 2015.8.4

    state_events
        The state_events option sends progress events as each function in
        a state run completes execution.

        .. versionadded:: 3006.0

    CLI Examples:

    .. code-block:: bash

        salt '*' state.highstate

        salt '*' state.highstate whitelist=sls1_to_run,sls2_to_run
        salt '*' state.highstate exclude=sls_to_exclude
        salt '*' state.highstate exclude="[{'id': 'id_to_exclude'}, {'sls': 'sls_to_exclude'}]"

        salt '*' state.highstate pillar="{foo: 'Foo!', bar: 'Bar!'}"
    """
    if _disabled(["highstate"]):
        log.debug(
            "Salt highstate run is disabled. To re-enable, run state.enable highstate"
        )
        ret = {
            "name": (
                "Salt highstate run is disabled. To re-enable, run state.enable"
                " highstate"
            ),
            "result": "False",
            "comment": "Disabled",
        }
        return ret

    conflict = _check_queue(queue, kwargs)
    if conflict is not None:
        return conflict

    orig_test = __opts__.get("test", None)
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    opts["test"] = _get_test_value(test, **kwargs)

    if "env" in kwargs:
        # "env" is not supported; Use "saltenv".
        kwargs.pop("env")

    if "saltenv" in kwargs:
        opts["saltenv"] = kwargs["saltenv"]

    if "pillarenv" in kwargs:
        opts["pillarenv"] = kwargs["pillarenv"]

    pillar_override = kwargs.get("pillar")
    pillar_enc = kwargs.get("pillar_enc")
    if (
        pillar_enc is None
        and pillar_override is not None
        and not isinstance(pillar_override, dict)
    ):
        raise SaltInvocationError(
            "Pillar data must be formatted as a dictionary, unless pillar_enc "
            "is specified."
        )

    if state_events is not None:
        opts["state_events"] = state_events

    try:
        st_ = salt.state.HighState(
            opts,
            pillar_override,
            kwargs.get("__pub_jid"),
            pillar_enc=pillar_enc,
            proxy=dict(__proxy__),
            context=dict(__context__),
            mocked=kwargs.get("mock", False),
            initial_pillar=_get_initial_pillar(opts),
        )
    except NameError:
        st_ = salt.state.HighState(
            opts,
            pillar_override,
            kwargs.get("__pub_jid"),
            pillar_enc=pillar_enc,
            mocked=kwargs.get("mock", False),
            initial_pillar=_get_initial_pillar(opts),
        )

    with st_:
        errors = _get_pillar_errors(kwargs, st_.opts["pillar"])
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            return ["Pillar failed to render with the following messages:"] + errors

        st_.push_active()
        orchestration_jid = kwargs.get("orchestration_jid")
        snapper_pre = _snapper_pre(opts, kwargs.get("__pub_jid", "called localy"))
        try:
            ret = st_.call_highstate(
                exclude=kwargs.get("exclude", []),
                cache=kwargs.get("cache", None),
                cache_name=kwargs.get("cache_name", "highstate"),
                force=kwargs.get("force", False),
                whitelist=kwargs.get("whitelist"),
                orchestration_jid=orchestration_jid,
            )
        finally:
            st_.pop_active()

        if isinstance(ret, dict) and (
            __salt__["config.option"]("state_data", "") == "terse"
            or kwargs.get("terse")
        ):
            ret = _filter_running(ret)

        _set_retcode(ret, highstate=st_.building_highstate)
        _snapper_post(opts, kwargs.get("__pub_jid", "called localy"), snapper_pre)

        # Work around Windows multiprocessing bug, set __opts__['test'] back to
        # value from before this function was run.
        __opts__["test"] = orig_test

        return ret


def sls(
    mods,
    test=None,
    exclude=None,
    queue=None,
    sync_mods=None,
    state_events=None,
    **kwargs,
):
    """
    Execute the states in one or more SLS files

    test
        Run states in test-only (dry-run) mode

    pillar
        Custom Pillar values, passed as a dictionary of key-value pairs

        .. code-block:: bash

            salt '*' state.sls stuff pillar='{"foo": "bar"}'

        .. note::
            Values passed this way will override existing Pillar values set via
            ``pillar_roots`` or an external Pillar source.  Pillar values that
            are not included in the kwarg will not be overwritten.

        .. versionchanged:: 2016.3.0
            GPG-encrypted CLI Pillar data is now supported via the GPG
            renderer. See :ref:`here <encrypted-cli-pillar-data>` for details.

    pillar_enc
        Specify which renderer to use to decrypt encrypted data located within
        the ``pillar`` value. Currently, only ``gpg`` is supported.

        .. versionadded:: 2016.3.0

    exclude
        Exclude specific states from execution. Accepts a list of sls names, a
        comma-separated string of sls names, or a list of dictionaries
        containing ``sls`` or ``id`` keys. Glob-patterns may be used to match
        multiple states.

        .. code-block:: bash

            salt '*' state.sls foo,bar,baz exclude=bar,baz
            salt '*' state.sls foo,bar,baz exclude=ba*
            salt '*' state.sls foo,bar,baz exclude="[{'id': 'id_to_exclude'}, {'sls': 'sls_to_exclude'}]"

    queue : False
        Instead of failing immediately when another state run is in progress,
        a value of ``True`` will queue the new state run to begin running once
        the other has finished.

        This option starts a new thread for each queued state run, so use this
        option sparingly.

        .. versionchanged:: 3006.0
            This parameter can also be set via the ``state_queue`` configuration
            option. Additionally, it can now be set to an integer representing
            the maximum queue size which can be attained before the state runs
            will fail to be queued. This can prevent runaway conditions where
            new threads are started until system performance is hampered.

    concurrent : False
        Execute state runs concurrently instead of serially

        .. warning::

            This flag is potentially dangerous. It is designed for use when
            multiple state runs can safely be run at the same time. Do *not*
            use this flag for performance optimization.

    saltenv
        Specify a salt fileserver environment to be used when applying states

        .. versionchanged:: 0.17.0
            Argument name changed from ``env`` to ``saltenv``.

        .. versionchanged:: 2014.7.0
            If no saltenv is specified, the minion config will be checked for an
            ``environment`` parameter and if found, it will be used. If none is
            found, ``base`` will be used. In prior releases, the minion config
            was not checked and ``base`` would always be assumed when the
            saltenv was not explicitly set.

    pillarenv
        Specify a Pillar environment to be used when applying states. This
        can also be set in the minion config file using the
        :conf_minion:`pillarenv` option. When neither the
        :conf_minion:`pillarenv` minion config option nor this CLI argument is
        used, all Pillar environments will be merged together.

    localconfig
        Optionally, instead of using the minion config, load minion opts from
        the file specified by this argument, and then merge them with the
        options from the minion config. This functionality allows for specific
        states to be run with their own custom minion configuration, including
        different pillars, file_roots, etc.

    mock
        The mock option allows for the state run to execute without actually
        calling any states. This then returns a mocked return which will show
        the requisite ordering as well as fully validate the state run.

        .. versionadded:: 2015.8.4

    sync_mods
        If specified, the desired custom module types will be synced prior to
        running the SLS files:

        .. code-block:: bash

            salt '*' state.sls stuff sync_mods=states,modules
            salt '*' state.sls stuff sync_mods=all

        .. versionadded:: 2017.7.8,2018.3.3,2019.2.0

    state_events
        The state_events option sends progress events as each function in
        a state run completes execution.

        .. versionadded:: 3006.0

    CLI Example:

    .. code-block:: bash

        # Run the states configured in salt://example.sls (or salt://example/init.sls)
        salt '*' state.apply example

        # Run the states configured in salt://core.sls (or salt://core/init.sls)
        # and salt://edit/vim.sls (or salt://edit/vim/init.sls)
        salt '*' state.sls core,edit.vim

        # Run the states configured in a more deeply nested directory such as salt://my/nested/state.sls (or salt://my/nested/state/init.sls)
        salt '*' state.sls my.nested.state

        salt '*' state.sls core exclude="[{'id': 'id_to_exclude'}, {'sls': 'sls_to_exclude'}]"
        salt '*' state.sls myslsfile pillar="{foo: 'Foo!', bar: 'Bar!'}"
    """
    concurrent = kwargs.get("concurrent", False)
    if "env" in kwargs:
        # "env" is not supported; Use "saltenv".
        kwargs.pop("env")

    # Modification to __opts__ lost after this
    conflict = _check_queue(queue, kwargs)
    if conflict is not None:
        return conflict

    if isinstance(mods, list):
        disabled = _disabled(mods)
    else:
        disabled = _disabled([mods])

    if disabled:
        for state in disabled:
            log.debug(
                "Salt state %s is disabled. To re-enable, run state.enable %s",
                state,
                state,
            )
        __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
        return disabled

    orig_test = __opts__.get("test", None)
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)

    opts["test"] = _get_test_value(test, **kwargs)

    # Since this is running a specific SLS file (or files), fall back to the
    # 'base' saltenv if none is configured and none was passed.
    if opts["saltenv"] is None:
        opts["saltenv"] = "base"

    pillar_override = kwargs.get("pillar")
    pillar_enc = kwargs.get("pillar_enc")
    if (
        pillar_enc is None
        and pillar_override is not None
        and not isinstance(pillar_override, dict)
    ):
        raise SaltInvocationError(
            "Pillar data must be formatted as a dictionary, unless pillar_enc "
            "is specified."
        )

    cfn = os.path.join(
        __opts__["cachedir"],
        "{}.cache.p".format(kwargs.get("cache_name", "highstate")),
    )

    if sync_mods is True:
        sync_mods = ["all"]
    if sync_mods is not None:
        sync_mods = salt.utils.args.split_input(sync_mods)
    else:
        sync_mods = []

    if "all" in sync_mods and sync_mods != ["all"]:
        # Prevent unnecessary extra syncing
        sync_mods = ["all"]

    for module_type in sync_mods:
        try:
            __salt__[f"saltutil.sync_{module_type}"](saltenv=opts["saltenv"])
        except KeyError:
            log.warning("Invalid custom module type '%s', ignoring", module_type)

    if state_events is not None:
        opts["state_events"] = state_events

    try:
        st_ = salt.state.HighState(
            opts,
            pillar_override,
            kwargs.get("__pub_jid"),
            pillar_enc=pillar_enc,
            proxy=dict(__proxy__),
            context=dict(__context__),
            mocked=kwargs.get("mock", False),
            initial_pillar=_get_initial_pillar(opts),
        )
    except NameError:
        st_ = salt.state.HighState(
            opts,
            pillar_override,
            kwargs.get("__pub_jid"),
            pillar_enc=pillar_enc,
            mocked=kwargs.get("mock", False),
            initial_pillar=_get_initial_pillar(opts),
        )

    with st_:
        errors = _get_pillar_errors(kwargs, pillar=st_.opts["pillar"])
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            return ["Pillar failed to render with the following messages:"] + errors

        orchestration_jid = kwargs.get("orchestration_jid")
        with salt.utils.files.set_umask(0o077):
            if kwargs.get("cache"):
                if os.path.isfile(cfn):
                    with salt.utils.files.fopen(cfn, "rb") as fp_:
                        high_ = salt.payload.load(fp_)
                        return st_.state.call_high(high_, orchestration_jid)

        # If the state file is an integer, convert to a string then to unicode
        if isinstance(mods, int):
            mods = salt.utils.stringutils.to_unicode(str(mods))

        mods = salt.utils.args.split_input(mods)

        st_.push_active()
        try:
            high_, errors = st_.render_highstate({opts["saltenv"]: mods})

            if errors:
                __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
                return errors

            if exclude:
                exclude = salt.utils.args.split_input(exclude)
                if "__exclude__" in high_:
                    high_["__exclude__"].extend(exclude)
                else:
                    high_["__exclude__"] = exclude
            snapper_pre = _snapper_pre(opts, kwargs.get("__pub_jid", "called localy"))
            ret = st_.state.call_high(high_, orchestration_jid)
        finally:
            st_.pop_active()
    if __salt__["config.option"]("state_data", "") == "terse" or kwargs.get("terse"):
        ret = _filter_running(ret)
    cache_file = os.path.join(__opts__["cachedir"], "sls.p")
    with salt.utils.files.set_umask(0o077):
        try:
            if salt.utils.platform.is_windows():
                # Make sure cache file isn't read-only
                __salt__["cmd.run"](["attrib", "-R", cache_file], python_shell=False)
            with salt.utils.files.fopen(cache_file, "w+b") as fp_:
                salt.payload.dump(ret, fp_)
        except OSError:
            log.error(
                "Unable to write to SLS cache file %s. Check permission.", cache_file
            )
        _set_retcode(ret, high_)
        # Work around Windows multiprocessing bug, set __opts__['test'] back to
        # value from before this function was run.
        __opts__["test"] = orig_test

        try:
            with salt.utils.files.fopen(cfn, "w+b") as fp_:
                try:
                    salt.payload.dump(high_, fp_)
                except TypeError:
                    # Can't serialize pydsl
                    pass
        except OSError:
            log.error(
                "Unable to write to highstate cache file %s. Do you have permissions?",
                cfn,
            )

    _snapper_post(opts, kwargs.get("__pub_jid", "called localy"), snapper_pre)
    return ret


def top(topfn, test=None, queue=None, **kwargs):
    """
    Execute a specific top file instead of the default. This is useful to apply
    configurations from a different environment (for example, dev or prod), without
    modifying the default top file.

    queue : False
        Instead of failing immediately when another state run is in progress,
        a value of ``True`` will queue the new state run to begin running once
        the other has finished.

        This option starts a new thread for each queued state run, so use this
        option sparingly.

        .. versionchanged:: 3006.0
            This parameter can also be set via the ``state_queue`` configuration
            option. Additionally, it can now be set to an integer representing
            the maximum queue size which can be attained before the state runs
            will fail to be queued. This can prevent runaway conditions where
            new threads are started until system performance is hampered.

    saltenv
        Specify a salt fileserver environment to be used when applying states

    pillarenv
        Specify a Pillar environment to be used when applying states. This
        can also be set in the minion config file using the
        :conf_minion:`pillarenv` option. When neither the
        :conf_minion:`pillarenv` minion config option nor this CLI argument is
        used, all Pillar environments will be merged together.

        .. versionadded:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' state.top reverse_top.sls
        salt '*' state.top prod_top.sls exclude=sls_to_exclude
        salt '*' state.top dev_top.sls exclude="[{'id': 'id_to_exclude'}, {'sls': 'sls_to_exclude'}]"
    """
    conflict = _check_queue(queue, kwargs)
    if conflict is not None:
        return conflict
    orig_test = __opts__.get("test", None)
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    opts["test"] = _get_test_value(test, **kwargs)

    pillar_override = kwargs.get("pillar")
    pillar_enc = kwargs.get("pillar_enc")
    if (
        pillar_enc is None
        and pillar_override is not None
        and not isinstance(pillar_override, dict)
    ):
        raise SaltInvocationError(
            "Pillar data must be formatted as a dictionary, unless pillar_enc "
            "is specified."
        )
    try:
        st_ = salt.state.HighState(
            opts,
            pillar_override,
            pillar_enc=pillar_enc,
            context=dict(__context__),
            proxy=dict(__proxy__),
            initial_pillar=_get_initial_pillar(opts),
        )
    except NameError:
        st_ = salt.state.HighState(
            opts,
            pillar_override,
            pillar_enc=pillar_enc,
            context=dict(__context__),
            initial_pillar=_get_initial_pillar(opts),
        )

    with st_:
        errors = _get_pillar_errors(kwargs, pillar=st_.opts["pillar"])
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            return ["Pillar failed to render with the following messages:"] + errors

        st_.push_active()
        st_.opts["state_top"] = salt.utils.url.create(topfn)
        ret = {}
        orchestration_jid = kwargs.get("orchestration_jid")
        if "saltenv" in kwargs:
            st_.opts["state_top_saltenv"] = kwargs["saltenv"]
        try:
            snapper_pre = _snapper_pre(opts, kwargs.get("__pub_jid", "called localy"))
            ret = st_.call_highstate(
                exclude=kwargs.get("exclude", []),
                cache=kwargs.get("cache", None),
                cache_name=kwargs.get("cache_name", "highstate"),
                orchestration_jid=orchestration_jid,
            )
        finally:
            st_.pop_active()

        _set_retcode(ret, highstate=st_.building_highstate)
        # Work around Windows multiprocessing bug, set __opts__['test'] back to
        # value from before this function was run.
        _snapper_post(opts, kwargs.get("__pub_jid", "called localy"), snapper_pre)
        __opts__["test"] = orig_test
        return ret


def show_highstate(queue=None, **kwargs):
    """
    Retrieve the highstate data from the salt master and display it

    Custom Pillar data can be passed with the ``pillar`` kwarg.

    CLI Example:

    .. code-block:: bash

        salt '*' state.show_highstate
    """
    conflict = _check_queue(queue, kwargs)
    if conflict is not None:
        return conflict
    pillar_override = kwargs.get("pillar")
    pillar_enc = kwargs.get("pillar_enc")
    if (
        pillar_enc is None
        and pillar_override is not None
        and not isinstance(pillar_override, dict)
    ):
        raise SaltInvocationError(
            "Pillar data must be formatted as a dictionary, unless pillar_enc "
            "is specified."
        )

    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    try:
        st_ = salt.state.HighState(
            opts,
            pillar_override,
            pillar_enc=pillar_enc,
            proxy=__proxy__,
            initial_pillar=_get_initial_pillar(opts),
        )
    except NameError:
        st_ = salt.state.HighState(
            opts,
            pillar_override,
            pillar_enc=pillar_enc,
            initial_pillar=_get_initial_pillar(opts),
        )

    with st_:
        errors = _get_pillar_errors(kwargs, pillar=st_.opts["pillar"])
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            raise CommandExecutionError("Pillar failed to render", info=errors)

        st_.push_active()
        try:
            ret = st_.compile_highstate()
        finally:
            st_.pop_active()
        _set_retcode(ret)
        return ret


def show_lowstate(queue=None, **kwargs):
    """
    List out the low data that will be applied to this minion

    CLI Example:

    .. code-block:: bash

        salt '*' state.show_lowstate
    """
    conflict = _check_queue(queue, kwargs)
    if conflict is not None:
        assert False
        return conflict

    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    try:
        st_ = salt.state.HighState(
            opts, proxy=__proxy__, initial_pillar=_get_initial_pillar(opts)
        )
    except NameError:
        st_ = salt.state.HighState(opts, initial_pillar=_get_initial_pillar(opts))

    with st_:
        errors = _get_pillar_errors(kwargs, pillar=st_.opts["pillar"])
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            raise CommandExecutionError("Pillar failed to render", info=errors)

        st_.push_active()
        try:
            ret = st_.compile_low_chunks()
        finally:
            st_.pop_active()
        return ret


def show_state_usage(queue=None, **kwargs):
    """
    Retrieve the highstate data from the salt master to analyse used and unused states

    Custom Pillar data can be passed with the ``pillar`` kwarg.

    CLI Example:

    .. code-block:: bash

        salt '*' state.show_state_usage
    """
    conflict = _check_queue(queue, kwargs)
    if conflict is not None:
        return conflict
    pillar = kwargs.get("pillar")
    pillar_enc = kwargs.get("pillar_enc")
    if pillar_enc is None and pillar is not None and not isinstance(pillar, dict):
        raise SaltInvocationError(
            "Pillar data must be formatted as a dictionary, unless pillar_enc "
            "is specified."
        )

    with salt.state.HighState(__opts__, pillar, pillar_enc=pillar_enc) as st_:
        st_.push_active()

        try:
            ret = st_.compile_state_usage()
        finally:
            st_.pop_active()
        _set_retcode(ret)
        return ret


def show_states(queue=None, **kwargs):
    """
    Returns the list of states that will be applied on highstate.

    CLI Example:

    .. code-block:: bash

        salt '*' state.show_states

    .. versionadded:: 2019.2.0

    """
    conflict = _check_queue(queue, kwargs)
    if conflict is not None:
        return conflict

    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    try:
        st_ = salt.state.HighState(
            opts, proxy=__proxy__, initial_pillar=_get_initial_pillar(opts)
        )
    except NameError:
        st_ = salt.state.HighState(opts, initial_pillar=_get_initial_pillar(opts))

    with st_:
        errors = _get_pillar_errors(kwargs, pillar=st_.opts["pillar"])
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            raise CommandExecutionError("Pillar failed to render", info=errors)

        st_.push_active()
        states = OrderedDict()
        try:
            result = st_.compile_low_chunks()

            if not isinstance(result, list):
                raise Exception(result)

            for s in result:
                if not isinstance(s, dict):
                    _set_retcode(result)
                    return result
                states[s["__sls__"]] = True
        finally:
            st_.pop_active()

        return list(states.keys())


def sls_id(id_, mods, test=None, queue=None, state_events=None, **kwargs):
    """
    Call a single ID from the named module(s) and handle all requisites

    The state ID comes *before* the module ID(s) on the command line.

    id
        ID to call

    mods
        Comma-delimited list of modules to search for given id and its requisites

    .. versionadded:: 2014.7.0

    saltenv : base
        Specify a salt fileserver environment to be used when applying states

    pillarenv
        Specify a Pillar environment to be used when applying states. This
        can also be set in the minion config file using the
        :conf_minion:`pillarenv` option. When neither the
        :conf_minion:`pillarenv` minion config option nor this CLI argument is
        used, all Pillar environments will be merged together.

    pillar
        Custom Pillar values, passed as a dictionary of key-value pairs

        .. code-block:: bash

            salt '*' state.sls_id my_state my_module pillar='{"foo": "bar"}'

        .. note::
            Values passed this way will override existing Pillar values set via
            ``pillar_roots`` or an external Pillar source.  Pillar values that
            are not included in the kwarg will not be overwritten.

        .. versionadded:: 2018.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' state.sls_id my_state my_module

        salt '*' state.sls_id my_state my_module,a_common_module
    """
    conflict = _check_queue(queue, kwargs)
    if conflict is not None:
        return conflict
    orig_test = __opts__.get("test", None)
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    opts["test"] = _get_test_value(test, **kwargs)

    # Since this is running a specific ID within a specific SLS file, fall back
    # to the 'base' saltenv if none is configured and none was passed.
    if opts["saltenv"] is None:
        opts["saltenv"] = "base"

    pillar_override = kwargs.get("pillar")
    pillar_enc = kwargs.get("pillar_enc")
    if (
        pillar_enc is None
        and pillar_override is not None
        and not isinstance(pillar_override, dict)
    ):
        raise SaltInvocationError(
            "Pillar data must be formatted as a dictionary, unless pillar_enc "
            "is specified."
        )

    if state_events is not None:
        opts["state_events"] = state_events

    try:
        st_ = salt.state.HighState(
            opts,
            pillar_override,
            pillar_enc=pillar_enc,
            proxy=__proxy__,
            initial_pillar=_get_initial_pillar(opts),
        )
    except NameError:
        st_ = salt.state.HighState(
            opts,
            pillar_override,
            pillar_enc=pillar_enc,
            initial_pillar=_get_initial_pillar(opts),
        )

    with st_:
        errors = _get_pillar_errors(kwargs, pillar=st_.opts["pillar"])
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            return ["Pillar failed to render with the following messages:"] + errors

        split_mods = salt.utils.args.split_input(mods)
        st_.push_active()
        try:
            high_, errors = st_.render_highstate({opts["saltenv"]: split_mods})
        finally:
            st_.pop_active()
        errors += st_.state.verify_high(high_)
        # Apply requisites to high data
        high_, req_in_errors = st_.state.requisite_in(high_)
        if req_in_errors:
            # This if statement should not be necessary if there were no errors,
            # but it is required to get the unit tests to pass.
            errors.extend(req_in_errors)
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
            return errors
        chunks = st_.state.compile_high_data(high_)
        ret = {}
        for chunk in chunks:
            if chunk.get("__id__", "") == id_:
                ret.update(st_.state.call_chunk(chunk, {}, chunks))

        _set_retcode(ret, highstate=highstate)
        # Work around Windows multiprocessing bug, set __opts__['test'] back to
        # value from before this function was run.
        __opts__["test"] = orig_test
        if not ret:
            raise SaltInvocationError(
                "No matches for ID '{}' found in SLS '{}' within saltenv '{}'".format(
                    id_, mods, opts["saltenv"]
                )
            )
        return ret


def show_low_sls(mods, test=None, queue=None, **kwargs):
    """
    Display the low data from a specific sls. The default environment is
    ``base``, use ``saltenv`` to specify a different environment.

    saltenv
        Specify a salt fileserver environment to be used when applying states

    pillar
        Custom Pillar values, passed as a dictionary of key-value pairs

        .. code-block:: bash

            salt '*' state.show_low_sls stuff pillar='{"foo": "bar"}'

        .. note::
            Values passed this way will override Pillar values set via
            ``pillar_roots`` or an external Pillar source.

    pillarenv
        Specify a Pillar environment to be used when applying states. This
        can also be set in the minion config file using the
        :conf_minion:`pillarenv` option. When neither the
        :conf_minion:`pillarenv` minion config option nor this CLI argument is
        used, all Pillar environments will be merged together.

    CLI Example:

    .. code-block:: bash

        salt '*' state.show_low_sls foo
        salt '*' state.show_low_sls foo saltenv=dev
    """
    if "env" in kwargs:
        # "env" is not supported; Use "saltenv".
        kwargs.pop("env")

    conflict = _check_queue(queue, kwargs)
    if conflict is not None:
        return conflict
    orig_test = __opts__.get("test", None)
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    opts["test"] = _get_test_value(test, **kwargs)

    # Since this is dealing with a specific SLS file (or files), fall back to
    # the 'base' saltenv if none is configured and none was passed.
    if opts["saltenv"] is None:
        opts["saltenv"] = "base"

    pillar_override = kwargs.get("pillar")
    pillar_enc = kwargs.get("pillar_enc")
    if (
        pillar_enc is None
        and pillar_override is not None
        and not isinstance(pillar_override, dict)
    ):
        raise SaltInvocationError(
            "Pillar data must be formatted as a dictionary, unless pillar_enc "
            "is specified."
        )

    try:
        st_ = salt.state.HighState(
            opts,
            pillar_override,
            proxy=__proxy__,
            initial_pillar=_get_initial_pillar(opts),
        )
    except NameError:
        st_ = salt.state.HighState(
            opts, pillar_override, initial_pillar=_get_initial_pillar(opts)
        )

    with st_:
        errors = _get_pillar_errors(kwargs, pillar=st_.opts["pillar"])
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            raise CommandExecutionError("Pillar failed to render", info=errors)

        mods = salt.utils.args.split_input(mods)
        st_.push_active()
        try:
            high_, errors = st_.render_highstate({opts["saltenv"]: mods})
        finally:
            st_.pop_active()
        errors += st_.state.verify_high(high_)
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
            return errors
        ret = st_.state.compile_high_data(high_)
        # Work around Windows multiprocessing bug, set __opts__['test'] back to
        # value from before this function was run.
        __opts__["test"] = orig_test
        return ret


def show_sls(mods, test=None, queue=None, **kwargs):
    """
    Display the state data from a specific sls or list of sls files on the
    master. The default environment is ``base``, use ``saltenv`` to specify a
    different environment.

    This function does not support topfiles.  For ``top.sls`` please use
    ``show_top`` instead.

    Custom Pillar data can be passed with the ``pillar`` kwarg.

    saltenv
        Specify a salt fileserver environment to be used when applying states

    pillarenv
        Specify a Pillar environment to be used when applying states. This
        can also be set in the minion config file using the
        :conf_minion:`pillarenv` option. When neither the
        :conf_minion:`pillarenv` minion config option nor this CLI argument is
        used, all Pillar environments will be merged together.

    CLI Example:

    .. code-block:: bash

        salt '*' state.show_sls core,edit.vim saltenv=dev
    """
    if "env" in kwargs:
        # "env" is not supported; Use "saltenv".
        kwargs.pop("env")

    conflict = _check_queue(queue, kwargs)
    if conflict is not None:
        return conflict
    orig_test = __opts__.get("test", None)
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)

    opts["test"] = _get_test_value(test, **kwargs)

    # Since this is dealing with a specific SLS file (or files), fall back to
    # the 'base' saltenv if none is configured and none was passed.
    if opts["saltenv"] is None:
        opts["saltenv"] = "base"

    pillar_override = kwargs.get("pillar")
    pillar_enc = kwargs.get("pillar_enc")
    if (
        pillar_enc is None
        and pillar_override is not None
        and not isinstance(pillar_override, dict)
    ):
        raise SaltInvocationError(
            "Pillar data must be formatted as a dictionary, unless pillar_enc "
            "is specified."
        )

    try:
        st_ = salt.state.HighState(
            opts,
            pillar_override,
            pillar_enc=pillar_enc,
            proxy=__proxy__,
            initial_pillar=_get_initial_pillar(opts),
        )
    except NameError:
        st_ = salt.state.HighState(
            opts,
            pillar_override,
            pillar_enc=pillar_enc,
            initial_pillar=_get_initial_pillar(opts),
        )

    with st_:
        errors = _get_pillar_errors(kwargs, pillar=st_.opts["pillar"])
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            raise CommandExecutionError("Pillar failed to render", info=errors)

        mods = salt.utils.args.split_input(mods)
        st_.push_active()
        try:
            high_, errors = st_.render_highstate({opts["saltenv"]: mods})
        finally:
            st_.pop_active()
        errors += st_.state.verify_high(high_)
        # Work around Windows multiprocessing bug, set __opts__['test'] back to
        # value from before this function was run.
        __opts__["test"] = orig_test
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
            return errors
        return high_


def sls_exists(mods, test=None, queue=None, **kwargs):
    """
    Tests for the existence the of a specific SLS or list of SLS files on the
    master. Similar to :py:func:`state.show_sls <salt.modules.state.show_sls>`,
    rather than returning state details, returns True or False. The default
    environment is ``base``, use ``saltenv`` to specify a different environment.

    .. versionadded:: 2019.2.0

    saltenv
        Specify a salt fileserver environment from which to look for the SLS files
        specified in the ``mods`` argument

    CLI Example:

    .. code-block:: bash

        salt '*' state.sls_exists core,edit.vim saltenv=dev
    """
    return isinstance(show_sls(mods, test=test, queue=queue, **kwargs), dict)


def id_exists(ids, mods, test=None, queue=None, **kwargs):
    """
    Tests for the existence of a specific ID or list of IDs within the
    specified SLS file(s). Similar to :py:func:`state.sls_exists
    <salt.modules.state.sls_exists>`, returns True or False. The default
    environment is base``, use ``saltenv`` to specify a different environment.

    .. versionadded:: 2019.2.0

    saltenv
        Specify a salt fileserver environment from which to look for the SLS files
        specified in the ``mods`` argument

    CLI Example:

    .. code-block:: bash

        salt '*' state.id_exists create_myfile,update_template filestate saltenv=dev
    """
    ids = salt.utils.args.split_input(ids)
    ids = set(ids)
    sls_ids = {
        x["__id__"] for x in show_low_sls(mods, test=test, queue=queue, **kwargs)
    }
    return ids.issubset(sls_ids)


def show_top(queue=None, **kwargs):
    """
    Return the top data that the minion will use for a highstate

    CLI Example:

    .. code-block:: bash

        salt '*' state.show_top
    """
    if "env" in kwargs:
        # "env" is not supported; Use "saltenv".
        kwargs.pop("env")

    conflict = _check_queue(queue, kwargs)
    if conflict is not None:
        return conflict

    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    try:
        st_ = salt.state.HighState(
            opts, proxy=__proxy__, initial_pillar=_get_initial_pillar(opts)
        )
    except NameError:
        st_ = salt.state.HighState(opts, initial_pillar=_get_initial_pillar(opts))

    with st_:
        errors = _get_pillar_errors(kwargs, pillar=st_.opts["pillar"])
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_PILLAR_FAILURE
            raise CommandExecutionError("Pillar failed to render", info=errors)

        errors = []
        top_ = st_.get_top()
        errors += st_.verify_tops(top_)
        if errors:
            __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
            return errors
        matches = st_.top_matches(top_)
        return matches


def single(fun, name, test=None, queue=None, **kwargs):
    """
    Execute a single state function with the named kwargs, returns False if
    insufficient data is sent to the command

    By default, the values of the kwargs will be parsed as YAML. So, you can
    specify lists values, or lists of single entry key-value maps, as you
    would in a YAML salt file. Alternatively, JSON format of keyword values
    is also supported.

    CLI Example:

    .. code-block:: bash

        salt '*' state.single pkg.installed name=vim

    """
    conflict = _check_queue(queue, kwargs)
    if conflict is not None:
        return conflict
    comps = fun.split(".")
    if len(comps) < 2:
        __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
        return "Invalid function passed"
    kwargs.update({"state": comps[0], "fun": comps[1], "__id__": name, "name": name})
    orig_test = __opts__.get("test", None)
    opts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    opts["test"] = _get_test_value(test, **kwargs)

    pillar_override = kwargs.get("pillar")
    pillar_enc = kwargs.get("pillar_enc")
    if (
        pillar_enc is None
        and pillar_override is not None
        and not isinstance(pillar_override, dict)
    ):
        raise SaltInvocationError(
            "Pillar data must be formatted as a dictionary, unless pillar_enc "
            "is specified."
        )

    try:
        st_ = salt.state.State(
            opts,
            pillar_override,
            pillar_enc=pillar_enc,
            proxy=__proxy__,
            initial_pillar=_get_initial_pillar(opts),
        )
    except NameError:
        st_ = salt.state.State(
            opts,
            pillar_override,
            pillar_enc=pillar_enc,
            initial_pillar=_get_initial_pillar(opts),
        )
    err = st_.verify_data(kwargs)
    if err:
        __context__["retcode"] = salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR
        return err

    st_._mod_init(kwargs)
    snapper_pre = _snapper_pre(opts, kwargs.get("__pub_jid", "called localy"))
    ret = {
        "{0[state]}_|-{0[__id__]}_|-{0[name]}_|-{0[fun]}".format(kwargs): st_.call(
            kwargs
        )
    }
    _set_retcode(ret)
    # Work around Windows multiprocessing bug, set __opts__['test'] back to
    # value from before this function was run.
    _snapper_post(opts, kwargs.get("__pub_jid", "called localy"), snapper_pre)
    __opts__["test"] = orig_test
    return ret


def clear_cache():
    """
    Clear out cached state files, forcing even cache runs to refresh the cache
    on the next state execution.

    Remember that the state cache is completely disabled by default, this
    execution only applies if cache=True is used in states

    CLI Example:

    .. code-block:: bash

        salt '*' state.clear_cache
    """
    ret = []
    for fn_ in os.listdir(__opts__["cachedir"]):
        if fn_.endswith(".cache.p"):
            path = os.path.join(__opts__["cachedir"], fn_)
            if not os.path.isfile(path):
                continue
            os.remove(path)
            ret.append(fn_)
    return ret


def pkg(pkg_path, pkg_sum, hash_type, test=None, **kwargs):
    """
    Execute a packaged state run, the packaged state run will exist in a
    tarball available locally. This packaged state
    can be generated using salt-ssh.

    CLI Example:

    .. code-block:: bash

        salt '*' state.pkg /tmp/salt_state.tgz 760a9353810e36f6d81416366fc426dc md5
    """
    # TODO - Add ability to download from salt master or other source
    popts = salt.utils.state.get_sls_opts(__opts__, **kwargs)
    if not os.path.isfile(pkg_path):
        return {}
    if not salt.utils.hashutils.get_hash(pkg_path, hash_type) == pkg_sum:
        return {}
    root = tempfile.mkdtemp()
    s_pkg = tarfile.open(pkg_path, "r:gz")
    # Verify that the tarball does not extract outside of the intended root
    members = s_pkg.getmembers()
    for member in members:
        if salt.utils.stringutils.to_unicode(member.path).startswith(
            (os.sep, f"..{os.sep}")
        ):
            return {}
        elif f"..{os.sep}" in salt.utils.stringutils.to_unicode(member.path):
            return {}
    s_pkg.extractall(root)  # nosec
    s_pkg.close()
    lowstate_json = os.path.join(root, "lowstate.json")
    with salt.utils.files.fopen(lowstate_json, "r") as fp_:
        lowstate = salt.utils.json.load(fp_)
    # Check for errors in the lowstate
    for chunk in lowstate:
        if not isinstance(chunk, dict):
            return lowstate
    pillar_json = os.path.join(root, "pillar.json")
    if os.path.isfile(pillar_json):
        with salt.utils.files.fopen(pillar_json, "r") as fp_:
            pillar_override = salt.utils.json.load(fp_)
    else:
        pillar_override = None

    roster_grains_json = os.path.join(root, "roster_grains.json")
    if os.path.isfile(roster_grains_json):
        with salt.utils.files.fopen(roster_grains_json, "r") as fp_:
            roster_grains = _format_cached_grains(salt.utils.json.load(fp_))

    if os.path.isfile(roster_grains_json):
        popts["grains"] = roster_grains
    popts["fileclient"] = "local"
    popts["file_roots"] = {}
    popts["test"] = _get_test_value(test, **kwargs)
    envs = os.listdir(root)
    for fn_ in envs:
        full = os.path.join(root, fn_)
        if not os.path.isdir(full):
            continue
        popts["file_roots"][fn_] = [full]
    st_ = salt.state.State(popts, pillar_override=pillar_override)
    snapper_pre = _snapper_pre(popts, kwargs.get("__pub_jid", "called localy"))
    ret = st_.call_chunks(lowstate)
    ret = st_.call_listen(lowstate, ret)
    try:
        shutil.rmtree(root)
    except OSError:
        pass
    _set_retcode(ret)
    _snapper_post(popts, kwargs.get("__pub_jid", "called localy"), snapper_pre)
    return ret


def disable(states):
    """
    Disable state runs.

    CLI Example:

    .. code-block:: bash

        salt '*' state.disable highstate

        salt '*' state.disable highstate,test.succeed_without_changes

    .. note::
        To disable a state file from running provide the same name that would
        be passed in a state.sls call.

        salt '*' state.disable bind.config

    """
    ret = {"res": True, "msg": ""}

    states = salt.utils.args.split_input(states)

    msg = []
    _disabled_state_runs = __salt__["grains.get"]("state_runs_disabled")
    if not isinstance(_disabled_state_runs, list):
        _disabled_state_runs = []

    _changed = False
    for _state in states:
        if _state in _disabled_state_runs:
            msg.append(f"Info: {_state} state already disabled.")
        else:
            msg.append(f"Info: {_state} state disabled.")
            _disabled_state_runs.append(_state)
            _changed = True

    if _changed:
        __salt__["grains.setval"]("state_runs_disabled", _disabled_state_runs)

    ret["msg"] = "\n".join(msg)

    # refresh the grains
    __salt__["saltutil.refresh_modules"]()

    return ret


def enable(states):
    """
    Enable state function or sls run

    CLI Example:

    .. code-block:: bash

        salt '*' state.enable highstate

        salt '*' state.enable test.succeed_without_changes

    .. note::
        To enable a state file from running provide the same name that would
        be passed in a state.sls call.

        salt '*' state.disable bind.config

    """
    ret = {"res": True, "msg": ""}

    states = salt.utils.args.split_input(states)
    log.debug("states %s", states)

    msg = []
    _disabled_state_runs = __salt__["grains.get"]("state_runs_disabled")
    if not isinstance(_disabled_state_runs, list):
        _disabled_state_runs = []

    _changed = False
    for _state in states:
        log.debug("_state %s", _state)
        if _state not in _disabled_state_runs:
            msg.append(f"Info: {_state} state already enabled.")
        else:
            msg.append(f"Info: {_state} state enabled.")
            _disabled_state_runs.remove(_state)
            _changed = True

    if _changed:
        __salt__["grains.setval"]("state_runs_disabled", _disabled_state_runs)

    ret["msg"] = "\n".join(msg)

    # refresh the grains
    __salt__["saltutil.refresh_modules"]()

    return ret


def list_disabled():
    """
    List the states which are currently disabled

    CLI Example:

    .. code-block:: bash

        salt '*' state.list_disabled
    """
    return __salt__["grains.get"]("state_runs_disabled")


def _disabled(funs):
    """
    Return messages for disabled states
    that match state functions in funs.
    """
    ret = []
    _disabled_state_runs = __salt__["grains.get"]("state_runs_disabled")
    if not isinstance(_disabled_state_runs, list):
        _disabled_state_runs = []
    for state in funs:
        for _state in _disabled_state_runs:
            if ".*" in _state:
                target_state = _state.split(".")[0]
                target_state = (
                    target_state + "."
                    if not target_state.endswith(".")
                    else target_state
                )
                if state.startswith(target_state):
                    err = (
                        'The state file "{0}" is currently disabled by "{1}", '
                        "to re-enable, run state.enable {1}.".format(
                            state,
                            _state,
                        )
                    )
                    ret.append(err)
                    continue
            else:
                if _state == state:
                    err = (
                        'The state file "{0}" is currently disabled, '
                        "to re-enable, run state.enable {0}.".format(
                            _state,
                        )
                    )
                    ret.append(err)
                    continue
    return ret


def event(
    tagmatch="*", count=-1, quiet=False, sock_dir=None, pretty=False, node="minion"
):
    r"""
    Watch Salt's event bus and block until the given tag is matched

    .. versionadded:: 2016.3.0
    .. versionchanged:: 2019.2.0
        ``tagmatch`` can now be either a glob or regular expression.

    This is useful for utilizing Salt's event bus from shell scripts or for
    taking simple actions directly from the CLI.

    Enable debug logging to see ignored events.

    :param tagmatch: the event is written to stdout for each tag that matches
        this glob or regular expression.
    :param count: this number is decremented for each event that matches the
        ``tagmatch`` parameter; pass ``-1`` to listen forever.
    :param quiet: do not print to stdout; just block
    :param sock_dir: path to the Salt master's event socket file.
    :param pretty: Output the JSON all on a single line if ``False`` (useful
        for shell tools); pretty-print the JSON output if ``True``.
    :param node: Watch the minion-side or master-side event bus.

    CLI Example:

    .. code-block:: bash

        salt-call --local state.event pretty=True
    """
    with salt.utils.event.get_event(
        node,
        sock_dir or __opts__["sock_dir"],
        opts=__opts__,
        listen=True,
    ) as sevent:

        while True:
            ret = sevent.get_event(full=True, auto_reconnect=True)
            if ret is None:
                continue

            if salt.utils.stringutils.expr_match(ret["tag"], tagmatch):
                if not quiet:
                    salt.utils.stringutils.print_cli(
                        "{}\t{}".format(
                            salt.utils.stringutils.to_str(ret["tag"]),
                            salt.utils.json.dumps(
                                salt.utils.data.decode(ret["data"]),
                                sort_keys=pretty,
                                indent=None if not pretty else 4,
                            ),
                        )
                    )
                    sys.stdout.flush()

                if count > 0:
                    count -= 1
                    log.debug("Remaining event matches: %s", count)

                if count == 0:
                    break
            else:
                log.debug("Skipping event tag: %s", ret["tag"])
                continue



"""
Pure python state renderer
==========================

To use this renderer, the SLS file should contain a function called ``run``
which returns highstate data.

The highstate data is a dictionary containing identifiers as keys, and execution
dictionaries as values. For example the following state declaration in YAML:

.. code-block:: yaml

    common_packages:
      pkg.installed:
       - pkgs:
          - curl
          - vim

translates to:

.. code-block:: python

     {'common_packages': {'pkg.installed': [{'pkgs': ['curl', 'vim']}]}}

In this module, a few objects are defined for you, giving access to Salt's
execution functions, grains, pillar, etc. They are:

- ``__salt__`` - :ref:`Execution functions <all-salt.modules>` (i.e.
  ``__salt__['test.echo']('foo')``)
- ``__grains__`` - :ref:`Grains <targeting-grains>` (i.e. ``__grains__['os']``)
- ``__pillar__`` - :ref:`Pillar data <pillar>` (i.e. ``__pillar__['foo']``)
- ``__opts__`` - Minion configuration options
- ``__env__`` - The effective salt fileserver environment (i.e. ``base``). Also
  referred to as a "saltenv". ``__env__`` should not be modified in a pure
  python SLS file. To use a different environment, the environment should be
  set when executing the state. This can be done in a couple different ways:

  - Using the ``saltenv`` argument on the salt CLI (i.e. ``salt '*' state.sls
    foo.bar.baz saltenv=env_name``).
  - By adding a ``saltenv`` argument to an individual state within the SLS
    file. In other words, adding a line like this to the state's data
    structure: ``{'saltenv': 'env_name'}``

- ``__sls__`` - The SLS path of the file. For example, if the root of the base
  environment is ``/srv/salt``, and the SLS file is
  ``/srv/salt/foo/bar/baz.sls``, then ``__sls__`` in that file will be
  ``foo.bar.baz``.

When used in a scenario where additional user-provided context data is supplied
(such as with :mod:`file.managed <salt.states.file.managed>`), the additional
data will typically be injected into the script as one or more global
variables:

.. code-block:: jinja

    /etc/http/conf/http.conf:
      file.managed:
        - source: salt://apache/generate_http_conf.py
        - template: py
        - context:
            # Will be injected as the global variable "site_name".
            site_name: {{ site_name }}

When writing a reactor SLS file the global context ``data`` (same as context
``{{ data }}`` for states written with Jinja + YAML) is available. The
following YAML + Jinja state declaration:

.. code-block:: jinja

    {% if data['id'] == 'mysql1' %}
    highstate_run:
      local.state.apply:
        - tgt: mysql1
    {% endif %}

translates to:

.. code-block:: python

    if data['id'] == 'mysql1':
        return {'highstate_run': {'local.state.apply': [{'tgt': 'mysql1'}]}}

Full Example
------------

.. code-block:: python
   :linenos:

    #!py

    def run():
        config = {}

        if __grains__['os'] == 'Ubuntu':
            user = 'ubuntu'
            group = 'ubuntu'
            home = '/home/{0}'.format(user)
        else:
            user = 'root'
            group = 'root'
            home = '/root/'

        config['s3cmd'] = {
            'pkg': [
                'installed',
                {'name': 's3cmd'},
            ],
        }

        config[home + '/.s3cfg'] = {
            'file.managed': [
                {'source': 'salt://s3cfg/templates/s3cfg'},
                {'template': 'jinja'},
                {'user': user},
                {'group': group},
                {'mode': 600},
                {'context': {
                    'aws_key': __pillar__['AWS_ACCESS_KEY_ID'],
                    'aws_secret_key': __pillar__['AWS_SECRET_ACCESS_KEY'],
                    },
                },
            ],
        }

        return config

"""

import os

import salt.utils.templates
from salt.exceptions import SaltRenderError


def render(template, saltenv="base", sls="", tmplpath=None, **kws):
    """
    Render the python module's components

    :rtype: string
    """
    template = tmplpath
    if not os.path.isfile(template):
        raise SaltRenderError(f"Template {template} is not a file!")

    tmp_data = salt.utils.templates.py(
        template,
        True,
        __salt__=__salt__,
        salt=__salt__,
        __grains__=__grains__,
        grains=__grains__,
        __opts__=__opts__,
        opts=__opts__,
        __pillar__=__pillar__,
        pillar=__pillar__,
        __env__=saltenv,
        saltenv=saltenv,
        __sls__=sls,
        sls=sls,
        **kws,
    )
    if not tmp_data.get("result", False):
        raise SaltRenderError(
            tmp_data.get("data", "Unknown render error in py renderer")
        )

    return tmp_data["data"]



"""
    :codeauthor: Gareth J. Greenaway <gareth@saltstack.com>
"""

import pytest

import salt.states.rsync as rsync
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {rsync: {}}


def test_syncronized_failed():
    """
    Test to perform an rsync.synchorized call that fails
    """
    ret = {"pid": 100, "retcode": 23, "stderr": "", "stdout": ""}
    _expected = {
        "changes": {},
        "comment": "Destination directory name was not found.",
        "name": "name",
        "result": False,
    }

    with patch.dict(rsync.__opts__, {"test": True}):
        mock = MagicMock(return_value=ret)
        with patch.dict(rsync.__salt__, {"rsync.rsync": mock}):
            assert rsync.synchronized("name", "source") == _expected

    # Run again mocking os.path.exists as True
    ret = {
        "pid": 100,
        "retcode": 23,
        "stderr": "Something went wrong",
        "stdout": "",
    }
    _expected = {
        "changes": {},
        "comment": "Something went wrong",
        "name": "name",
        "result": False,
    }

    with patch("os.path.exists", MagicMock(return_value=True)):
        with patch.dict(rsync.__opts__, {"test": False}):
            mock = MagicMock(return_value=ret)
            with patch.dict(rsync.__salt__, {"rsync.rsync": mock}):
                assert rsync.synchronized("name", "source") == _expected


def test_syncronized():
    """
    Test to perform an rsync.synchorized call that succeeds
    """
    ret = {
        "pid": 100,
        "retcode": 0,
        "stderr": "",
        "stdout": (
            "sending incremental file list\nsnapshot.jar\n\n"
            "sent 106 bytes  received 35 bytes  282.00 bytes/sec\n"
            "total size is 0  speedup is 0.00"
        ),
    }
    _expected = {
        "changes": {"copied": "snapshot.jar", "deleted": "N/A"},
        "comment": (
            "- sent 106 bytes\n- received 35 bytes\n- "
            "282.00 bytes/sec\n- total size is 0\n- speedup is 0.00"
        ),
        "name": "name",
        "result": True,
    }

    with patch("os.path.exists", MagicMock(return_value=True)):
        with patch.dict(rsync.__opts__, {"test": False}):
            mock = MagicMock(return_value=ret)
            with patch.dict(rsync.__salt__, {"rsync.rsync": mock}):
                assert rsync.synchronized("name", "source") == _expected

    # Second pass simulating the file being in place
    ret = {
        "pid": 100,
        "retcode": 0,
        "stderr": "",
        "stdout": (
            "sending incremental file list\n\n"
            "sent 106 bytes  received 35 bytes  "
            "282.00 bytes/sec\ntotal size is 0  "
            "speedup is 0.00"
        ),
    }
    _expected = {
        "changes": {},
        "comment": (
            "- sent 106 bytes\n- received "
            "35 bytes\n- 282.00 bytes/sec\n- total "
            "size is 0\n- speedup is 0.00"
        ),
        "name": "name",
        "result": True,
    }

    with patch("os.path.exists", MagicMock(return_value=True)):
        with patch.dict(rsync.__opts__, {"test": False}):
            mock = MagicMock(return_value=ret)
            with patch.dict(rsync.__salt__, {"rsync.rsync": mock}):
                assert rsync.synchronized("name", "source") == _expected


def test_syncronized_test_true():
    """
    Test to perform an rsync.synchorized call that fails
    """
    ret = {
        "pid": 100,
        "retcode": 23,
        "stderr": "Something went wrong",
        "stdout": "",
    }
    _expected = {"changes": {}, "comment": "- ", "name": "name", "result": None}

    with patch("os.path.exists", MagicMock(return_value=True)):
        with patch.dict(rsync.__opts__, {"test": True}):
            mock = MagicMock(return_value=ret)
            with patch.dict(rsync.__salt__, {"rsync.rsync": mock}):
                assert rsync.synchronized("name", "source") == _expected



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)


    salt.utils.odict
    ~~~~~~~~~~~~~~~~

    Implements a DefaultOrderedDict Class that serves as a
    combination of ``OrderedDict`` and ``defaultdict``
    Its source was submitted here::

        http://stackoverflow.com/questions/6190331/
"""

import copy
from collections import OrderedDict
from collections.abc import Callable

from salt.utils.versions import warn_until


class DefaultOrderedDict(OrderedDict):
    """
    Dictionary that remembers insertion order
    """

    def __init__(self, default_factory=None, *a, **kw):
        warn_until(
            3009,
            "This module is deprecated. Use the standard library's collections.OrderedDict "
            "or salt.utils.datastructures instead.",
        )
        if default_factory is not None and not isinstance(default_factory, Callable):
            raise TypeError("first argument must be callable")
        super().__init__(*a, **kw)
        self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        if self.default_factory is None:
            args = tuple()
        else:
            args = (self.default_factory,)
        return type(self), args, None, None, self.items()

    def copy(self):
        return copy.copy(self)

    def __copy__(self):
        return type(self)(self.default_factory, self)

    def __deepcopy__(self):
        import copy

        return type(self)(self.default_factory, copy.deepcopy(self.items()))

    def __repr__(self, _repr_running={}):  # pylint: disable=W0102
        return "DefaultOrderedDict({}, {})".format(
            self.default_factory, super().__repr__()
        )


class HashableOrderedDict(OrderedDict):
    def __init__(self):
        warn_until(
            3009,
            "This module is deprecated. Use the standard library's collections.OrderedDict "
            "or salt.utils.datastructures instead.",
        )
        super().__init__()

    def __hash__(self):
        return id(self)



import logging

import pytest

from tests.conftest import FIPS_TESTRUN

log = logging.getLogger(__name__)


@pytest.fixture(scope="package")
def master(request, salt_factories):
    config_defaults = {
        "transport": request.config.getoption("--transport"),
    }
    config_overrides = {
        "interface": "127.0.0.1",
        "auto_accept": True,
        "gather_job_timeout": 60,
        "timeout": 60,
        "order_masters": True,
        "master_id": "master",
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
    }
    factory = salt_factories.salt_master_daemon(
        "master",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    with factory.started(start_timeout=180):
        yield factory


@pytest.fixture(scope="package")
def salt_cli(master):
    return master.salt_cli(timeout=180)


@pytest.fixture(scope="package")
def syndic(master, salt_factories):

    ret_port = master.config["ret_port"]
    port = master.config["publish_port"]
    addr = master.config["interface"]

    # Force both master's publish port to be the same, this is a drawback of
    # the current syndic design.
    config_defaults = {
        "transport": master.config["transport"],
        "interface": "127.0.0.2",
        "publish_port": f"{port}",
        "syndic_mode": "cluster",
    }
    master_overrides = {
        "interface": "127.0.0.2",
        "auto_accept": True,
        "syndic_master": f"{addr}",
        "syndic_master_port": f"{ret_port}",
        "master_id": "syndic",
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
    }
    minion_overrides = {
        "master": "127.0.0.2",
        "publish_port": f"{port}",
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }
    factory = master.salt_syndic_daemon(
        "syndic",
        defaults=config_defaults,
        master_overrides=master_overrides,
        minion_overrides=minion_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    with factory.started(start_timeout=180):
        yield factory


@pytest.fixture(scope="package")
def minion(syndic, salt_factories):
    config_defaults = {
        "transport": syndic.config["transport"],
    }
    port = syndic.master.config["ret_port"]
    addr = syndic.master.config["interface"]
    config_overrides = {
        "master": f"{addr}:{port}",
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }
    factory = syndic.master.salt_minion_daemon(
        "minion",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    with factory.started(start_timeout=180):
        yield factory
